import os
import pathlib
import tempfile

import h5py
import numpy as np

from .. import halo, units, util
from .gadgethdf import GadgetHDFSnap, _GadgetHdfMultiFileManager


class SwiftMultiFileManager(_GadgetHdfMultiFileManager):

    def __init__(self, filename: pathlib.Path, take_cells, take_region, mode='r'):
        self._take_cells = take_cells

        if take_cells is not None and take_region is not None:
            raise ValueError("Either take_cells or take_region must be specified, not both")

        if take_cells is not None or take_region is not None:
            # we need to avoid loading a VDS file, as it won't have the info needed to make our own VDS
            # pointers to the right cells
            try:
                with h5py.File(filename, mode='r') as f:
                    if f['Header'].attrs['Virtual'][0] == 1 and filename.suffix == '.hdf5':
                        # if we simply strip off .hdf5, GadgetHDFSnap will find the underlying files (hopefully..)
                        filename = filename.with_suffix('')
                    # it's not a virtual file anyway, so we're ok
            except FileNotFoundError:
                pass # perfect, this is either going to be pieced together by pynbody, or it'll fail anyway
        super().__init__(filename, mode)

        if take_region is not None:
            # convert take to take_cells
            take_cells = self._identify_cells_to_take(take_region)

        # hopefully the shenanigans above mean we don't end up with a VDS, but just in case we do, check again:
        if self.is_virtual() and take_cells is not None:
            raise ValueError("Can't take a subset of cells from a VDS-based HDF5 file pointing to multiple underlying files")

        self._make_hdf_vfile(take_cells)

    def _identify_cells_to_take(self, take):
        centres = self[0]['Cells/Centres'][:]
        return np.where(take.cubic_cell_intersection(centres))[0]

    def get_unit_attrs(self):
        return self[0].parent['InternalCodeUnits'].attrs

    def get_header_attrs(self):
        return self[0].parent['Parameters'].attrs

    def is_virtual(self):
        return self[0].parent['Header'].attrs['Virtual'][0] == 1

    def iter_particle_groups_with_name(self, hdf_family_name):
        if hdf_family_name in self._hdf_vfile:
            if self._size_from_hdf5_key in self._hdf_vfile[hdf_family_name]:
                yield self._hdf_vfile[hdf_family_name]

    def _all_group_names(self):
        return self[0]['Cells/Counts'].keys()

    def _make_hdf_vfile(self, take_cells):
        with tempfile.TemporaryDirectory() as tmpdirname:
            tmpfile_path = os.path.join(tmpdirname,"nofile.hdf5")
            with h5py.File(name=tmpfile_path, mode='w') as hdf_vfile:
                # ideally one would simply use  backing_store=False, to File but then there doesn't seem to be a way
                # to actually use the file (the VDS views just returns zeros).
                # Instead we write then re-read it, which presumably carries minimal overhead but is a bit ugly.

                for group_name in self._all_group_names():

                    if take_cells is None:
                        source_groups, source_slices = self._generate_groups_and_slices_for_full_file(group_name)
                    else:
                        source_groups, source_slices = self._generate_groups_and_slices_from_cells(group_name,
                                                                                                   take_cells)

                    target_group = hdf_vfile.create_group(group_name)
                    self._make_hdf_group_with_slicing(source_groups, source_slices, target_group)

            self._hdf_vfile = h5py.File(name=tmpfile_path, mode='r')
            # on exiting the temporarydirectory context, the file/folder will be unlinked but we should
            # be able to retain access for the lifetime of the hdf_vfile object

    def _generate_groups_and_slices_for_full_file(self, group_name):
        source_groups = [hdf_file[group_name] for hdf_file in self]
        source_lens = [len(f[group_name][self._size_from_hdf5_key]) for f in self]
        source_slices = [[slice(0, l)] for l in source_lens]
        return source_groups, source_slices

    def _generate_groups_and_slices_from_cells(self, group_name, take_cells):
        # we get the cell info from the first file, and assume it's consistent between files
        offsets = self[0]['Cells']['OffsetsInFile'][group_name]
        lens = self[0]['Cells']['Counts'][group_name]
        file_responsible = self[0]['Cells']['Files'][group_name]

        # First, make a map from the file ID to the slices to be taken from that file
        file_id_to_slices_map = {}
        for cell in take_cells:
            if file_responsible[cell] not in file_id_to_slices_map:
                file_id_to_slices_map[file_responsible[cell]] = []
            sl = slice(offsets[cell], offsets[cell] + lens[cell])
            file_id_to_slices_map[file_responsible[cell]].append(sl)

        # Now, reformat everything into the format expected by _make_hdf_group_with_slicing
        source_slices = []
        source_groups = []

        for file_id in sorted(list(file_id_to_slices_map)):
            source_groups.append(self[file_id][group_name])
            source_slices.append(sorted(file_id_to_slices_map[file_id]))

        return source_groups, source_slices

    def _make_hdf_group_with_slicing(self, source_groups, source_slices, target_group):
        total_len = 0
        for take_slices in source_slices:
            for s in take_slices:
                assert s.step is None
                total_len += s.stop - s.start

        for array_name in source_groups[0]:
            offset = 0
            target_dims = (total_len,) + source_groups[0][array_name].shape[1:]
            layout = h5py.VirtualLayout(shape=target_dims, dtype=source_groups[0][array_name].dtype)

            for source_group, take_slices in zip(source_groups, source_slices):
                source = h5py.VirtualSource(source_group[array_name])
                for s in take_slices:
                    layout[slice(offset, offset+s.stop-s.start)] = source[s]
                    offset += s.stop-s.start

            target_group.create_virtual_dataset(array_name, layout)


class ExtractScalarWrapper:
    def __init__(self, underlying):
        self.underlying = underlying

    def __getitem__(self, name):
        val = self.underlying[name]
        try:
            return val[0]
        except (TypeError, IndexError):
            return val

class SwiftSnap(GadgetHDFSnap):
    _multifile_manager_class = SwiftMultiFileManager
    _readable_hdf5_test_key = "Policy"

    _velocity_unit_key = None
    _length_unit_key = 'Unit length in cgs (U_L)'
    _mass_unit_key = 'Unit mass in cgs (U_M)'
    _time_unit_key = 'Unit time in cgs (U_t)'

    _namemapper_config_section = 'swift-name-mapping'

    def __init__(self, filename, take_swift_cells=None, take_region=None):
        self._take_swift_cells = take_swift_cells
        self._take_region = take_region
        super().__init__(filename)


    def _init_hdf_filemanager(self, filename):
        self._hdf_files = self._multifile_manager_class(filename, self._take_swift_cells, self._take_region)

    def _is_cosmological(self):
        cosmo = ExtractScalarWrapper(self._hdf_files[0]['Cosmology'].attrs)
        return cosmo['Cosmological run'] == 1
    def _init_properties(self):
        params = ExtractScalarWrapper(self._hdf_files[0]['Parameters'].attrs)
        header = ExtractScalarWrapper(self._hdf_files[0]['Header'].attrs)
        cosmo = ExtractScalarWrapper(self._hdf_files[0]['Cosmology'].attrs)

        cosmological = self._is_cosmological()

        assert header['Dimension'] == 3, "Sorry, pynbody is only set up to deal with 3-dimensional swift simulations"

        if cosmological:
            self.properties['z'] = 1./(1.+cosmo['Scale-factor'])
            self.properties['a'] = cosmo['Scale-factor']
            self.properties['h'] = cosmo['h']
            # TODO: check these params are OK even at higher redshift (sample file is z=0)
            self.properties['omegaM0'] = cosmo['Omega_m']
            self.properties['omegaL0'] = cosmo['Omega_lambda']
            self.properties['omegaB0'] = cosmo['Omega_b']
            self.properties['omegaC0'] = cosmo['Omega_cdm']
            self.properties['omegaNu0'] = cosmo['Omega_nu_0']

            self.properties['boxsize'] = header['BoxSize']*self.infer_original_units('m')
            # Swift writes out 3D box sizes. Check it's actually a cube and if not emit a warning
            boxsize_3d = header.underlying['BoxSize']
            assert np.allclose(boxsize_3d[0], boxsize_3d)

        self.properties['time'] = header['Time']*self.infer_original_units("s")


    def _get_units_from_hdf_attr(self, hdfattrs):
        this_unit = units.Unit("1")
        for k in hdfattrs.keys():
            if k.endswith('exponent'):
                unitname = k.split(" ")[0]
                exponent = util.fractions.Fraction.from_float(float(ExtractScalarWrapper(hdfattrs)[k])).limit_denominator()

                if exponent != 0:
                    this_unit *= self._hdf_unitvar.get(unitname, units.no_unit)**exponent

        return this_unit

    def _init_unit_information(self):
        atr = ExtractScalarWrapper(self._hdf_files.get_unit_attrs())
        dist_unit = atr['Unit length in cgs (U_L)'] * units.cm
        mass_unit = atr['Unit mass in cgs (U_M)'] * units.g
        time_unit = atr['Unit time in cgs (U_t)'] * units.s
        temp_unit = atr['Unit temperature in cgs (U_T)'] * units.K
        vel_unit = dist_unit / time_unit

        unitvar = {'U_V': vel_unit,
                   'U_L': dist_unit,
                   'U_M': mass_unit,
                   'U_t': time_unit,
                   'U_T': temp_unit,
                   'a-scale': units.a,
                   'h-scale': units.h}


        self._hdf_unitvar = unitvar

        if self._is_cosmological():
            # when using hdf-provided exponents, the scalefactor will automatically be included. However, if using
            # infer_original_units, our best guess is that distances are comoving. So include the scalefactor
            # in self._file_units_system but not in self._hdf_unitvar.
            dist_unit = dist_unit * units.a

        self._file_units_system = [vel_unit, dist_unit, mass_unit, temp_unit]

    def halos(self, **kwargs):
        h = super().halos(**kwargs)

        if isinstance(h, halo.number_array.HaloNumberCatalogue):
            ignore = int(self._hdf_files.get_parameter_attrs()['FOF:group_id_default'])
            return halo.number_array.HaloNumberCatalogue(self, ignore=ignore, **kwargs)

        return h



"""HBT+ halo catalogue support.

.. _hbt_plus_parent_groups:

HBT+ and parent groups
----------------------

HBT+ identifies halos, not the parent groups (using the SubFind terminology). As a result, it may be used alongside a
parent group catalogue. The *pynbody* reader can present a unified interface to the combined hierarchy, presenting
groups from the parent catalogue as the top-level objects, with the HBT+ halos as children of these groups, and
HBT+ subhalos as children of the HBT+ halos.

For example:

.. ipython::

  In [1]: import pynbody

  In [2]: s = pynbody.load('testdata/gadget4_subfind_HBT/snapshot_034.hdf5')

  In [3]: hbtplus_halos = s.halos(priority=["HBTPlusCatalogue"])

  In [4]: subfind_groups = s.halos(priority=["Gadget4SubfindHDFCatalogue"])

  In [5]: combined_catalogue = hbtplus_halos.with_groups_from(subfind_groups)

  In [6]: combined_catalogue[0]
  Out[6]: <SimSnap "testdata/gadget4_subfind_HBT/snapshot_034.hdf5:halo_0" len=307386>

We can tell that this is actually a SubFind group by inspecting its properties:

.. ipython::

  In [7]: combined_catalogue[0].properties['GroupMass'] # <-- a SubFind property
  Out[7]: Unit("7.80e+44 g h**-1")

  In [8]: subfind_groups[0].properties['GroupMass'] # <-- the same property accessed directly from SubFind
  Out[8]: Unit("7.80e+44 g h**-1")

The ``subhalos`` attribute of each halo in the ``combined_catalogue`` will return the HBT+ subhalos that are children of
the corresponding SubFind group:

.. ipython::

  In [14]: combined_catalogue[0].subhalos[0].properties['TrackId'] # <-- an HBT+ property
  Out[14]: 54

Naturally, not only the properties but the particle information is available in the combined catalogue.


"""
from __future__ import annotations

import pathlib
import re

import h5py
import numpy as np
from numpy.typing import NDArray

from . import HaloCatalogue, HaloParticleIndices
from .details import number_mapping


class HBTPlusCatalogue(HaloCatalogue):
    """A class to represent a HBT+ halo catalogue."""
    def __init__(self, sim, halo_numbers=None, filename=None):
        """Initialize a HBTPlusCatalogue object.

        Parameters
        ----------
        sim : SimSnap
            The simulation snapshot to which this catalogue applies.

        halo_numbers : str, optional
            How to number the halos. If None (default), use a zero-based indexing.

            * If ``track``, use the TrackId from the catalogue.
            * If ``length-order``, order by Nbound (descending), similar to the AHF option of the same name.

        filename : str, optional
            The filename of the HBTPlus catalogue. If the file is spanned across multiple outputs (e.g.
            ``path/to/SubSnap_034.0.hdf5``, ``SubSnap_034.1.hdf5``, etc.), pass the filename of the first file or
            the common prefix (``path/to/SubSnap_034``). If None (default), attempt to find the file automatically.
        """
        if filename is None:
            filename = self._infer_hbt_filename(sim)
        else:
            filename = self._map_user_filename_to_file_0(filename)

        self._file = h5py.File(filename, 'r')

        num_halos = int(self._file["NumberOfSubhalosInAllFiles"][0])
        if int(self._file["NumberOfFiles"][0])>1:
            from ..util import hdf_vds
            filename = str(filename)[:-7]
            all_files = [f"{filename}.{num}.hdf5" for num in range(int(self._file["NumberOfFiles"][0]))]
            maker = hdf_vds.HdfVdsMaker(all_files)
            self._file = maker.get_temporary_hdf_vfile()

        self._trackid_number_mapper = number_mapping.create_halo_number_mapper(self._file["Subhalos"]["TrackId"])

        if halo_numbers is None:
            number_mapper = number_mapping.SimpleHaloNumberMapper(0, num_halos)
        elif halo_numbers == 'track':
            number_mapper = self._trackid_number_mapper
        elif halo_numbers == 'length-order':
            osort = np.argsort(-self._file["Subhalos"]["Nbound"][:], kind='stable')
            number_mapper = number_mapping.NonMonotonicHaloNumberMapper(osort, ordering=True, start_index=0)
        else:
            raise ValueError(f"Invalid value for halo_numbers: {halo_numbers}")
        super().__init__(sim, number_mapper)

        self._setup_parents()

    def __del__(self):
        if hasattr(self, "_file"):
            self._file.close()

    @classmethod
    def _infer_hbt_filename(cls, sim):
        sim_filename: pathlib.Path  = sim.filename
        try:
            snap_num = int(re.search(r'_(\d+)', sim_filename.name).group(1))
        except AttributeError:
            raise FileNotFoundError(f'Could not infer HBT filename from {sim_filename}. '
                                    f'Try passing hbt_filename explicitly.') from None

        candidate_paths = [sim_filename.with_name(f'SubSnap_{snap_num:03d}.0.hdf5'),
                            sim_filename.parent / f'{snap_num:03d}' / f'SubSnap_{snap_num:03d}.0.hdf5']

        for candidate_path in candidate_paths:
            if candidate_path.exists():
                return candidate_path

        raise FileNotFoundError(f'Could not find HBTPlus catalogue for {sim_filename}. Try passing hbt_filename explicitly.')

    @classmethod
    def _map_user_filename_to_file_0(cls, filename):
        filename = pathlib.Path(filename)
        if not filename.exists():
            dot_hdf5 = filename.parent / (filename.name + '.hdf5')
            dot_0_hdf5 = filename.parent / (filename.name + '.0.hdf5')
            if dot_hdf5.exists():
                filename = dot_hdf5
            elif dot_0_hdf5.exists():
                filename = dot_0_hdf5
        return filename

    def _setup_parents(self):
        parents = np.empty(len(self), dtype=np.intp)
        parents.fill(-1)
        for i in range(len(self)):
            for child in self._trackid_number_mapper.number_to_index(self._file["NestedSubhalos"][i]):
                parents[child] = self.number_mapper.index_to_number(i)

        self._parents = parents

    def _map_trackid_to_pynbody_number(self, trackid: NDArray[int]) -> NDArray[int]:
        if self.number_mapper is self._trackid_number_mapper:
            return trackid
        else:
            return self.number_mapper.index_to_number(
                self._trackid_number_mapper.number_to_index(
                  trackid
                )
            )

    def _get_particle_indices_one_halo(self, halo_number) -> NDArray[int]:
        self._init_iord_to_fpos()

        iords = self._file["SubhaloParticles"][self.number_mapper.number_to_index(halo_number)]
        return np.sort(self._iord_to_fpos.map_ignoring_order(iords))

    def _get_all_particle_indices(self) -> HaloParticleIndices | tuple[np.ndarray, np.ndarray]:
        self._init_iord_to_fpos()
        indices = np.empty(self._file["Subhalos"]["Nbound"].sum(),
                           dtype=self._file["SubhaloParticles"][0].dtype)
        boundaries = np.empty((len(self._file["SubhaloParticles"]), 2),
                              dtype=np.intp)
        start = 0
        for i, halo_parts in enumerate(self._file['SubhaloParticles']):
            end = start + len(halo_parts)
            indices[start:end] = np.sort(self._iord_to_fpos.map_ignoring_order(halo_parts))
            boundaries[i] = (start,end)
            start = end
        return indices, boundaries

    def with_groups_from(self, other: HaloCatalogue) -> HaloCatalogue:
        """Return a new catalogue that combines an HBT+ halo catalogue with a parent group catalogue.

        For example:

        >>> grps = hbt_cat.with_groups_from(subfind_cat)
        >>> grps[0] # -> returns subfind_cat[0]
        >>> grps[0].properties['children'] # -> returns relevant halo numbers in hbt_cat
        >>> grps[0].subhalos # -> returns HBT+ subhalos that are children of subfind_cat[0]

        See also :ref:`hbt_plus_parent_groups`.
        """
        return HBTPlusCatalogueWithGroups(self, other)

    def get_properties_one_halo(self, halo_number) -> dict:
        index = self.number_mapper.number_to_index(halo_number)
        result = {}
        for k in self._file["Subhalos"].dtype.names:
            result[k] = self._file["Subhalos"][k][index]
        result['children'] = self._get_children_one_halo(index)
        result['parent'] = self._parents[index]
        return result

    def get_properties_all_halos(self, with_units=True) -> dict:
        result = {}
        for k in self._file["Subhalos"].dtype.names:
            result[k] = np.asarray(self._file["Subhalos"][k])
        result['children'] = [self._get_children_one_halo(i) for i in range(len(self))]
        result['parent'] = self._parents
        return result

    def _get_children_one_halo(self, index) -> NDArray[int]:
        return self._map_trackid_to_pynbody_number(self._file["NestedSubhalos"][index])



    @classmethod
    def _can_load(cls, sim, halo_numbers=None, filename=None):
        if filename is not None:
            filename = cls._map_user_filename_to_file_0(filename)
        try:
            hbt_filename = filename or cls._infer_hbt_filename(sim)
            if h5py.is_hdf5(hbt_filename):
                with h5py.File(hbt_filename, 'r') as f:
                    if "NumberOfFiles" in f:
                        return True
        except OSError:
            pass
        return False




class HBTPlusCatalogueWithGroups(HaloCatalogue):
    """A class to represent a HBT+ halo catalogue with parent groups from another finder.

    For more information about parent groups, see :ref:`hbt_plus_parent_groups`.
    """

    def __init__(self, hbt_cat: HBTPlusCatalogue, group_cat: HaloCatalogue):
        if hbt_cat.base is not group_cat.base:
            raise ValueError("The two catalogues must have the same base simulation snapshot")
        if not isinstance(hbt_cat, HBTPlusCatalogue):
            raise ValueError("The first catalogue must be a HBTPlusCatalogue")

        self._hbt_cat = hbt_cat
        self._group_cat = group_cat
        self._hbt_host_groups = np.asarray(hbt_cat._file["Subhalos"]["HostHaloId"])
        if self._hbt_host_groups.max() >= len(group_cat):
            raise ValueError("The HBT+ catalogue contains host groups that are not in the group catalogue")
        self._children = [[] for _ in range(len(group_cat))]
        for hbt_index, hbt_number in enumerate(hbt_cat.number_mapper):
            self._children[self._hbt_host_groups[hbt_index]].append(hbt_number)
        self._children = [np.array(c) for c in self._children]

        super().__init__(group_cat.base, group_cat.number_mapper)

    def load_all(self):
        self._hbt_cat.load_all()
        self._group_cat.load_all()

    def _get_particle_indices_one_halo(self, halo_number) -> NDArray[int]:
        return self._group_cat._get_particle_indices_one_halo(halo_number)

    def get_properties_one_halo(self, halo_number) -> dict:
        group_properties = self._group_cat.get_properties_one_halo(halo_number)
        group_properties['children'] = self._children[self.number_mapper.number_to_index(halo_number)]
        return group_properties

    def get_properties_all_halos(self, with_units=True) -> dict:
        properties_all = self._group_cat.get_properties_all_halos(with_units)
        properties_all['children'] = self._children
        return properties_all

    def _get_subhalo_catalogue(self, halo_number):
        index = self._group_cat.number_mapper.number_to_index(halo_number)
        return self._hbt_cat[self._children[index]]



"""

profile
=======

"""

import logging

import numpy as np
import pylab as p

from .. import config, filt, units
from ..analysis import angmom, halo, profile

logger = logging.getLogger('pynbody.plot.profile')


def rotation_curve(sim, center=True, r_units='kpc',
                   v_units='km s^-1', disk_height='100 pc', nbins=50,
                   bin_spacing='equaln', clear=True, quick=False,
                   filename=None, min=False, max=False, yrange=False,
                   legend=False, parts=False, axes=False, **kwargs):
    """

    Centre on potential minimum, align so that the disk is in the
    x-y plane, then use the potential in that plane to generate and
    plot a rotation curve.

    **needs documentation/description of the keyword arguments**

    """
    import pylab as p

    if center:
        angmom.faceon(sim)

    if min:
        min_r = min
    else:
        min_r = sim['rxy'].min()
    if max:
        max_r = max
    else:
        max_r = sim['rxy'].max()

    pro = profile.Profile(sim, type=bin_spacing, nbins=nbins,
                          rmin =min_r, rmax =max_r)

    r = pro['rbins'].in_units(r_units)
    if quick:
        v = pro['rotation_curve_spherical'].in_units(v_units)
    else:
        v = pro['v_circ'].in_units(v_units)

    if axes:
        p = axes
    else:
        import pylab as p
        if clear:
            p.clf()

    if parts:
        p.plot(r, v, label='total', **kwargs)
        gpro = profile.Profile(sim.gas, type=bin_spacing, nbins=nbins,
                               rmin =min_r, rmax =max_r)
        dpro = profile.Profile(sim.dark, type=bin_spacing, nbins=nbins,
                               rmin =min_r, rmax =max_r)
        spro = profile.Profile(sim.star, type=bin_spacing, nbins=nbins,
                               rmin =min_r, rmax =max_r)
        if quick:
            gv = gpro['rotation_curve_spherical'].in_units(v_units)
            dv = dpro['rotation_curve_spherical'].in_units(v_units)
            sv = spro['rotation_curve_spherical'].in_units(v_units)
        else:
            gv = gpro['v_circ'].in_units(v_units)
            dv = dpro['v_circ'].in_units(v_units)
            sv = spro['v_circ'].in_units(v_units)
        p.plot(r, gv, "--", label="gas")
        p.plot(r, dv, label="dark")
        p.plot(r, sv, linestyle="dotted", label="star")
    else:
        p.plot(r, v, **kwargs)

    if yrange:
        p.axis(
            [min_r, units.Unit(max_r).in_units(r.units), yrange[0], yrange[1]])

    if not axes:
        p.xlabel("r / $" + r.units.latex() + "$", fontsize='large')
        p.ylabel("v$_c / " + v.units.latex() + '$', fontsize='large')

    if legend:
        p.legend(loc=0)

    if (filename):
        logger.info("Saving %s", filename)
        p.savefig(filename)

    return r, v


def fourier_profile(sim, center=True, disk_height='2 kpc', nbins=50,
                    pretime='2 Gyr', r_units='kpc', bin_spacing='equaln',
                    clear=True, min=False, max=False, filename=None, **kwargs):
    """
    Centre on potential minimum, align so that the disk is in the
    x-y plane, then plot the amplitude of the 2nd fourier mode as a
    function of radius.

    **needs description of the keyword arguments**

    """

    if center:
        angmom.faceon(sim)

    if min:
        min_r = min
    else:
        min_r = sim['rxy'].min()
    if max:
        max_r = max
    else:
        max_r = sim['rxy'].max()

    if isinstance(pretime, str):
        pretime = units.Unit(pretime)

    diskstars = sim.star[filt.Disc(max_r, disk_height)]
    youngstars = np.where(diskstars['tform'].in_units("Myr") >
                          sim.properties['time'].in_units(
                              "Myr", **sim.conversion_context())
                          - pretime.in_units('Myr'))[0]

    pro = profile.Profile(diskstars[youngstars], type=bin_spacing,
                          nbins=nbins, rmin =min_r, rmax =max_r)

    r = pro['rbins'].in_units(r_units)
    fourierprof = pro['fourier']
    a2 = fourierprof['amp'][2]

    if clear:
        p.clf()

    p.plot(r, a2, **kwargs)

    p.xlabel("r / $" + r.units.latex() + "$")
    p.ylabel("Amplitude of 2nd Fourier Mode")
    if (filename):
        logger.info("Saving %s", filename)
        p.savefig(filename)


def density_profile(sim, linestyle=False, center=True, clear=True, fit=False,in_units=None,
                    filename=None, fit_factor=0.02, axes=False, **kwargs):
    '''

    3d density profile

    **Options:**

    *filename* (None):  name of file to which to save output

    **Usage:**

    >>> import pynbody.plot as pp
    >>> h = s.halos()
    >>> pp.density_profile(h[1],linestyle='dashed',color='k')


    '''
    if axes: plt = axes
    else: import matplotlib.pyplot as plt

    global config

    logger.info("Centering...")
    if center:
        halo.center(sim, mode='ssc')

    logger.info("Creating profile...")

    if 'min' in kwargs:
        ps = profile.Profile(
            sim, ndim=3, type='log', nbins=40, rmin =kwargs['min'])
        del kwargs['min']
    else:
        ps = profile.Profile(sim, ndim=3, type='log', nbins=40)

    if clear and not axes:
        plt.clf()
    critden = (units.Unit('100 km s^-1 Mpc^-1')
               * sim.properties['h']) ** 2 / 8.0 / np.pi / units.G
    r = ps['rbins'].in_units('kpc')
    if in_units is None:
        den = ps['density'].in_units(critden)
    else:
        den = ps['density'].in_units(in_units)

    if linestyle:
        plt.errorbar(r, den, yerr=den / np.sqrt(ps['n']),
                     linestyle=linestyle, **kwargs)
    else:
        plt.errorbar(r, den, yerr=den / np.sqrt(ps['n']),
                     fmt='o', **kwargs)

    if in_units is None:
        ylabel=r'$\rho / \rho_{cr}$'  # +den.units.latex()+'$]')
    else:
        ylabel=r'$\rho / '+den.units.latex()+'$'
    if axes:
        plt.set_yscale('log')
        plt.set_xscale('log')
        plt.set_xlabel('r [kpc]')
        plt.set_ylabel(ylabel) #r'$\rho / \rho_{cr}$')  # +den.units.latex()+'$]')
    else:
        plt.yscale('log')
        plt.xscale('log')
        plt.xlabel('r [kpc]')
        plt.ylabel(ylabel) #r'$\rho / \rho_{cr}$')  # +den.units.latex()+'$]')


    if (filename):
        logger.info("Saving %s", filename)
        plt.savefig(filename)

    if fit:
        fit_inds = np.where(r < fit_factor*sim['r'].max())
        alphfit = np.polyfit(np.log10(r[fit_inds]),
                             np.log10(den[fit_inds]), 1)

        fit = np.poly1d(alphfit)
        plt.plot(r[fit_inds], 10**fit(np.log10(r[fit_inds])),
                 color='k',linestyle='dashed',
                 label=r'$\alpha$=%.1f'%alphfit[0])
        plt.legend(loc=3)

        return alphfit[0]



import numpy as np
import pytest

import pynbody
import pynbody.snapshot


def test_copy_on_access_subsnap_data_isolation():
    # Test the copy_on_access subsnap, which only gets data from the underlying rather than pointing back to it
    # This is used by tangos to implement shared memory server mode
    f = pynbody.new(10)
    f['blob'] = np.arange(10)

    for subscript in ([2, 3, 4], slice(2,5)):

        f_sub = f[subscript].get_copy_on_access_simsnap()
        # doesn't simply copy everything in to start with:
        assert 'blob' not in f_sub.keys()

        # can get the underlying data lazily
        assert (f_sub['blob'] == [2,3,4]).all()

        f_sub['blob'] = [100,101,102]
        assert (f_sub['blob'] == [100, 101, 102]).all()

        # copy_on_access: shouldn't have updated the underlying
        assert (f['blob'] == np.arange(10)).all()


class ExampleSnap(pynbody.snapshot.simsnap.SimSnap):
    pass


@ExampleSnap.derived_array
def foo(sim):
    return sim['blob']+5


def test_copy_on_access_subsnap_emulating_class():
    f = pynbody.new(10, class_=ExampleSnap)
    f['blob'] = np.arange(10)

    f_sub = f[[2, 3, 4]].get_copy_on_access_simsnap()
    assert (f_sub['foo'] == [7, 8, 9]).all()

    assert 'foo' not in f.keys()

def test_copy_on_access_subsnap_emulating_class_two_layers_down():
    f = pynbody.new(10, class_=ExampleSnap)
    f['blob'] = np.arange(10)

    f_sub_copy = f[[2, 3, 4]].get_copy_on_access_simsnap()

    f_sub_copy_copy = f_sub_copy.get_copy_on_access_simsnap()


    assert (f_sub_copy_copy['foo'] == [7, 8, 9]).all()

    assert 'foo' not in f_sub_copy.keys()
    assert 'foo' not in f.keys()



def test_copy_on_access_subsnap_family_array():
    f = pynbody.new(dm=10,star=10)
    f.dm['dm_only'] = np.arange(10)
    f.st['star_only'] = np.arange(10,20)

    f_sub = f[np.arange(0,20,2)].get_copy_on_access_simsnap()
    assert (f_sub.dm['dm_only'] == np.arange(0,10,2)).all()
    assert (f_sub.star['star_only'] == np.arange(10, 20, 2)).all()
    assert 'dm_only' not in f_sub.keys()
    assert 'star_only' not in f_sub.keys()

def test_copy_on_access_mixing_derived_and_nonderived():
    f = pynbody.new(dm=10, star=10, gas=10, class_=ExampleSnap)
    f.dm['blob'] = np.arange(10)
    f.star['foo'] = np.arange(10, 20)

    f_copy = f.get_copy_on_access_simsnap()
    assert (f_copy.dm['foo'] == np.arange(5, 15)).all()
    assert f_copy.dm['foo'].derived

    assert (f_copy.star['foo'] == np.arange(10, 20)).all()

    assert f_copy.dm['foo'].derived
    assert not f_copy.star['foo'].derived

    f_copy.dm['blob'] += 1

    assert (f_copy.dm['blob'] == np.arange(1, 11)).all()
    assert (f_copy.dm['foo'] == np.arange(6, 16)).all()
    assert (f_copy.star['foo'] == np.arange(10, 20)).all()
    assert (f.dm['blob'] == np.arange(10)).all()
def test_base_correctness():
    f = pynbody.new(10)
    f_sub = f[[2,3,4]]
    assert f_sub.ancestor is f

    f_sub = f[::2]
    assert f_sub.ancestor is f

    f_sub = f[[2,3,4]].get_copy_on_access_simsnap()
    assert f_sub.ancestor is f_sub
    assert not hasattr(f_sub, 'base')

def test_properties():
    f = pynbody.new(10)
    f.properties['test_property'] = 101
    f_c = f.get_copy_on_access_simsnap()
    # should have been copied in:
    assert f_c.properties['test_property'] == 101

    # should not reflect back to parent:
    f_c.properties['test_property'] = 100
    assert f_c.properties['test_property'] == 100
    assert f.properties['test_property'] == 101

def test_repr():
    f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
    f_c = f.get_copy_on_access_simsnap()

    assert repr(f_c) == '<SimSnap "testdata/gasoline_ahf/g15784.lr.01024:copied_on_access" len=1717156>'

def test_loadable_keys():
    f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
    f['pos'] # noqa
    f.dm['new_array'] = np.empty(len(f.dm))

    f_c = f.get_copy_on_access_simsnap()
    # anything loadable in the base is loadable in the copy
    assert 'pos' in f_c.loadable_keys()
    assert 'HI' in f_c.gas.loadable_keys()

    # check HI will actually load:
    f_c.gas['HI'] # noqa

    assert 'new_array' not in f_c.dm.keys()
    # it's in the parent keys, but not yet copied across
    assert 'new_array' in f_c.dm.loadable_keys()

def test_all_keys():
    f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
    f_c = f.get_copy_on_access_simsnap()
    assert 'pos' in f_c.all_keys()


def test_only_try_loading_once():
    f = pynbody.new(10)
    f_c = f.get_copy_on_access_simsnap()


    with pytest.raises(OSError, match="Not found"):
        f_c.dm._load_array('nonexistent')

    with pytest.raises(OSError, match="Previously tried"):
        f_c.dm._load_array('nonexistent')



"""AHF (Amiga Halo Finder) support"""

from __future__ import annotations

import glob
import gzip
import os.path
import pathlib
import re
import warnings

import numpy as np

from .. import snapshot, util
from . import HaloCatalogue, HaloParticleIndices, logger
from .details.number_mapping import (
    NonMonotonicHaloNumberMapper,
    SimpleHaloNumberMapper,
    create_halo_number_mapper,
)


class AHFCatalogue(HaloCatalogue):
    """
    Class to handle catalogues produced by Amiga Halo Finder (AHF).
    """

    def __init__(self, sim, filename=None, make_grp=None, get_all_parts=None, use_iord=None, ahf_basename=None,
                 dosort=None, only_stat=None, write_fpos=True, halo_numbers='file-order',
                 ignore_missing_substructure=True,
                 **kwargs):
        """Initialize an AHFCatalogue.

        Parameters
        ----------

        *sim*: SimSnap
          the simulation snapshot to which this catalogue refers

        *filename*: str | pathlib.Path
          specify a path to an AHF halo catalog. Note that AHF actually outputs multiple files; you can specify the
          path to the ``AHF_halos`` or ``AHF_particles`` file and the code will infer the other filenames from this.
          Alternatively you can specify the path up to the ``AHF_`` prefix and the code will similarly infer the full
          set of filenames.

        *halo_numbers*: str
          specify how to number the halos. Options are:

          * 'ahf' (default): use the halo numbers written in the AHF halos file if present, or a zero-based indexing
          * 'file-order': zero-based indexing of the halos
          * 'v1': one-based indexing of the halos, compatible with pynbody v1 default behaviour
          * 'length-order': sort by the number of particles in each halo, with the halo with most particles being halo 0
          * 'length-order-v1': as length-order, but indexing from halo 1, compatible with ``dosort=True`` in pynbody v1

        *ignore_missing_substructure*: bool
          If True (default), the code will not raise an exception if the substructure file is missing or corrupt. If
          False, it will raise an exception.

        *use_iord*: bool
          if True, the particle IDs in the Amiga catalogue are taken to refer to the iord array. If False,
          they are the particle offsets within the file. If None, the parameter defaults to True for GadgetSnap,
          False otherwise.

        *ahf_basename*: str
          Deprecated way to specify the location of the catalogue

        *make_grp*:
          Deprecated. If True a 'grp' array is created in the underlying snapshot specifying the lowest level halo
          that any given particle belongs to. If it is False, no such array is created; if None, the behaviour is
          determined by the configuration system.

        *get_all_parts*:
          Deprecated; use the :meth:`load_all` method instead.

        *dosort*:
          Deprecated; equivalent to ``halo_numbers='length-order'``

        *only_stat*:
          Deprecated; this keyword is now ignored. To obtain halo information without loading the particles,
          use the methods :meth:`get_properties_one_halo` or :meth:`get_properties_all_halos`.

        """

        if use_iord is None:
            use_iord = isinstance(
                sim.ancestor,
                (snapshot.gadget.GadgetSnap, snapshot.gadgethdf.GadgetHDFSnap)
            )

        self._use_iord = use_iord
        self._only_stat = only_stat
        self._try_writing_fpos = write_fpos

        if only_stat:
            warnings.warn(DeprecationWarning("only_stat keyword is deprecated; instead, use the catalogue's get_dummy_halo method"))

        if filename is not None:
            self._ahfBasename = str(self._user_specified_filename_to_ahf_basename(filename))
        elif ahf_basename is not None:
            warnings.warn(DeprecationWarning("ahf_basename keyword is deprecated; use filename instead"))
            self._ahfBasename = ahf_basename
        else:
            self._ahfBasename = self._infer_ahf_basename(sim)

        self._determine_format_revision_from_filename()

        logger.info("AHFCatalogue loading halo properties")
        self._load_ahf_halo_properties(self._ahfBasename + 'halos')

        if dosort:
            warnings.warn(DeprecationWarning("dosort keyword is deprecated; instead pass halo_numbers='length-order-v1'"))
            halo_numbers = 'length-order-v1'

        number_mapper = self._setup_halo_numbering(halo_numbers)

        super().__init__(sim, number_mapper)

        self._remap_host_halo_property()

        if self._use_iord:
            self._init_iord_to_fpos()

        try:
            self._load_ahf_substructure(self._ahfBasename + 'substructure')
        except (KeyError, ValueError, FileNotFoundError):
            if not ignore_missing_substructure:
                raise
            logger.error("Unable to load AHF substructure file; continuing without. To expose the underlying problem as an exception, pass ignore_missing_substructure=False to the AHFCatalogue constructor")

        if make_grp:
            warnings.warn(DeprecationWarning("make_grp keyword is deprecated; instead, use the catalogue's get_group_array method"))
            self.base['grp'] = self.get_group_array()

        if get_all_parts:
            warnings.warn(
                DeprecationWarning("get_all_parts keyword is deprecated; instead, use the catalogue's load_all method"))
            self.load_all()

        logger.info("AHFCatalogue loaded")

    def _setup_halo_numbering(self, halo_numbers):
        has_id = 'ID' in self._halo_properties
        if has_id:
            self._ahf_own_number_mapper = create_halo_number_mapper(self._halo_properties['ID'])
        else:
            # if no explicit IDs, ahf implicitly numbers starting at 0 in file order
            self._ahf_own_number_mapper = SimpleHaloNumberMapper(0, self._num_halos)

        if halo_numbers == 'v1':
            number_mapper = SimpleHaloNumberMapper(1, self._num_halos)
        elif halo_numbers == 'file-order':
            number_mapper = SimpleHaloNumberMapper(0, self._num_halos)
        elif halo_numbers == 'length-order' or halo_numbers == 'length-order-v1':
            npart = self._halo_properties['npart']
            osort = np.argsort(-npart, kind='stable')  # this is better than argsort(npart)[::-1], because it's stable
            if halo_numbers.endswith('v1'):
                number_mapper = NonMonotonicHaloNumberMapper(osort, ordering=True, start_index=1)
            else:
                number_mapper = NonMonotonicHaloNumberMapper(osort, ordering=True, start_index=0)
        elif halo_numbers == 'ahf':
            number_mapper = self._ahf_own_number_mapper
        else:
            raise ValueError(f"halo_numbers keyword {halo_numbers} not recognised")

        return number_mapper

    def _remap_host_halo_property(self):
        """When loaded from the .halos file, hostHalo is the AHF halo number of the host halo.
        This maps it onto whatever halo number pynbody is using."""

        if 'hostHalo' in self._halo_properties:
            host_halo = self._halo_properties['hostHalo']
            mask = host_halo != -1
            host_halo[mask] = self.number_mapper.index_to_number(
                self._ahf_own_number_mapper.number_to_index(host_halo[mask])
            )

    def _determine_format_revision_from_filename(self):
        if self._ahfBasename.split("z")[-2][-1] == ".":
            self._is_new_format = True
        else:
            self._is_new_format = False

    @classmethod
    def _infer_ahf_basename(cls, sim):
        candidates = cls._list_candidate_ahf_basenames(sim)
        if len(candidates) == 1:
            candidate = candidates[0]
        elif len(candidates) == 0:
            raise FileNotFoundError("No candidate AHF catalogue found; try specifying a catalogue using the "
                                    "ahf_basename keyword")
        else:
            candidate = candidates[0]
            warnings.warn(f"Multiple candidate AHF catalogues found; using {candidate}. To specify a different "
                          "catalogue, use the ahf_basename keyword, or move the other catalogues.")
        return util.cutgz(candidate)[:-9]


    def _write_fpos(self):
        try:
            np.savetxt(self._ahfBasename+'fpos', self._fpos, fmt="%d")
        except OSError:
            warnings.warn("Unable to write AHF_fpos file; performance will be reduced. Pass write_fpos=False to halo constructor to suppress this message, or use load_all() method to remove need for storing fpos information.")
        self._try_writing_fpos = False

    @property
    def base(self):
        return self._base()

    def _get_file_positions(self):
        """Get the starting positions of each halo's particle information within the
        AHF_particles file for faster access later"""
        if not hasattr(self, "_fpos"):
            if os.path.exists(self._ahfBasename + 'fpos'):
                self._fpos = np.loadtxt(self._ahfBasename+'fpos', dtype=int)
            else:
                self._fpos = np.empty(len(self.number_mapper), dtype=int)
                with util.open_(self._ahfBasename + 'particles') as f:
                    nhalo = int(f.readline().strip())
                    assert nhalo == len(self.number_mapper)
                    for hnum in range(nhalo):
                        npart = int(f.readline().strip())
                        assert npart == self._halo_properties['npart'][hnum]
                        self._fpos[hnum] = f.tell()
                        for i in range(npart):
                            f.readline()
                if self._try_writing_fpos:
                    if not os.path.exists(self._ahfBasename + 'fpos'):
                        self._write_fpos()

        return self._fpos

    def _load_ahf_particle_block(self, f, nparts):
        """Load the particles for the next halo described in particle file f"""
        if self._is_new_format:
            if not isinstance(f, gzip.GzipFile):
                data = np.fromfile(
                    f,
                    dtype=int,
                    sep=" ",
                    count=nparts * 2
                )[::2]
                data = np.ascontiguousarray(data)
            else:
                # unfortunately with gzipped files there does not
                # seem to be an efficient way to load nparts lines
                data = np.empty(nparts, dtype=int)
                for i in range(nparts):
                    data[i] = int(f.readline().split()[0])

            data = self._ahf_to_pynbody_particle_ids(data)
        else:
            if not isinstance(f, gzip.GzipFile):
                data = np.fromfile(f, dtype=int, sep=" ", count=nparts)
            else:
                # see comment above on gzipped files
                data = np.empty(nparts, dtype=int)
                for i in range(nparts):
                    data[i] = int(f.readline())
        data.sort()
        return data

    def _ahf_to_pynbody_particle_ids(self, data):
        ng = len(self.base.gas)
        nd = len(self.base.dark)
        ns = len(self.base.star)
        nds = nd + ns
        if self._use_iord:
            data = self._iord_to_fpos.map_ignoring_order(data)
        elif isinstance(self.base, snapshot.ramses.RamsesSnap):
            # AHF only expects three families, DM, star, gas in this order
            # and generates iords on disc according to this rule
            # For classical Ramses snapshots, this is perfectly adequate, but
            # for more modern outputs that have extra tracers, BHs families
            # we need to offset the ids to return the correct slicing
            # TODO These tests on snapshot type might not be necessary
            #  as using the family logic properly should be general.
            #  It is currently kept to ensure 100% backwards compatibility with previous behaviour,
            #  as this code is not explicitly checked by the pynbody test distribution

            if len(self.base) != nd + ns + ng:  # We have extra families to the base ones
                # First identify DM, star and gas particles in AHF
                ahf_dm_mask = data < nd
                ahf_star_mask = (data >= nd) & (data < nds)
                ahf_gas_mask = data >= nds

                # Then offset them by DM family start, to account for
                # additional families before it, e.g. gas tracers
                data[np.where(ahf_dm_mask)] += self.base._get_family_slice('dm').start

                # Star ids used to start at NDM, now they start with the star family slice
                offset = self.base._get_family_slice('star').start - nd
                data[np.where(ahf_star_mask)] += offset

                # Gas ids were greater than NDM + NSTAR, now they start with the gas slice
                offset = self.base._get_family_slice('gas').start - nds
                data[np.where(ahf_gas_mask)] += offset
        elif not isinstance(self.base, snapshot.nchilada.NchiladaSnap):
            hi_mask = data >= nds
            data[np.where(hi_mask)] -= nds
            data[np.where(~hi_mask)] += ng
        else:
            st_mask = (data >= nd) & (data < nds)
            g_mask = data >= nds
            data[np.where(st_mask)] += ng
            data[np.where(g_mask)] -= ns
        return data

    def _get_particle_indices_one_halo(self, halo_number):
        fpos = self._get_file_positions()
        file_index = self.number_mapper.number_to_index(halo_number)
        with util.open_(self._ahfBasename + 'particles') as f:
            f.seek(fpos[file_index],0)
            ids = self._load_ahf_particle_block(f, nparts=self._halo_properties['npart'][file_index])
        return ids

    def _get_all_particle_indices(self):
        boundaries = np.cumsum(np.concatenate(([0], self._halo_properties['npart'])))
        boundaries = np.vstack((boundaries[:-1], boundaries[1:])).T
        particle_ids = np.empty(boundaries[-1,1], dtype=int)
        with util.open_(self._ahfBasename + 'particles') as f:
            f.readline()
            for nparts, (start, end) in zip(self._halo_properties['npart'], boundaries):
                f.readline()
                particle_ids[start:end] = self._load_ahf_particle_block(f, nparts=nparts)


        return HaloParticleIndices(particle_ids=particle_ids, boundaries=boundaries)


    def get_properties_one_halo(self, i):
        index = self.number_mapper.number_to_index(i)
        return {key: self._halo_properties[key][index] for key in self._halo_properties}

    def get_properties_all_halos(self, with_units=True) -> dict:
        return self._halo_properties

    def _load_ahf_halo_properties(self, filename):
        # Note: we need to open in 'rt' mode in case the AHF catalogue
        # is gzipped.
        with util.open_(filename, "rt") as f:
            first_line = f.readline()
            lines = f.readlines()

        # get all the property names from the first, commented line
        # remove (#)
        fields = first_line.replace("#", "").split()
        keys = [re.sub(r'\([0-9]*\)', '', field)
                for field in fields]

        omit_first_column = False

        if self._is_new_format:
            # fix for column 0 being a non-column in some versions of the AHF
            # output
            if keys[0] == '':
                omit_first_column = True
                keys = keys[1:]

        self._halo_properties = {k: [] for k in keys}

        self._num_halos = len(lines)

        for line in lines:
            values = [
                float(x) if any(_ in x for _ in (".", "e", "nan", "inf"))
                else int(x)
                for x in line.split()
            ]
            # XXX Unit issues!  AHF uses distances in Mpc/h, possibly masses as
            # well
            if omit_first_column:
                values = values[1:]

            for key, value in zip(keys, values):
                self._halo_properties[key].append(value)

        for key in keys:
            self._halo_properties[key] = np.array(self._halo_properties[key])

    def _load_ahf_substructure(self, filename):
        with util.open_(filename) as f:
            lines = f.readlines()
        logger.info("AHFCatalogue loading substructure")

        if len(lines[0].split())!=2:
            # some variants of the format start with a line count, which we ignore
            lines = lines[1:]

        self._halo_properties['children'] = children = [[] for _ in range(self._num_halos)]
        self._halo_properties['parent'] = parent = np.empty(self._num_halos, dtype=int)
        parent.fill(-1)

        for halo_line, children_line in zip(lines[::2], lines[1::2]):
            haloid, _nsubhalos = (int(x) for x in halo_line.split())
            halo_index = self._ahf_own_number_mapper.number_to_index(haloid)
            halo_number = self.number_mapper.index_to_number(halo_index)
            children_ahf_numbering = [int(x) for x in children_line.split()]
            children_index = self._ahf_own_number_mapper.number_to_index(children_ahf_numbering)
            children_pynbody_numbering = self.number_mapper.index_to_number(children_index)

            children[halo_index] = children_pynbody_numbering
            for child_index in children_index:
                parent[child_index] = halo_number


    @staticmethod
    def _list_candidate_ahf_basenames(sim):
        candidates = set(glob.glob(f"{sim._filename}*z*particles*"))
        # use a set to ensure that no duplicates can be produced
        # This could arise in an edge case where _filename is a directory
        # and having the "/" at the end of it would lead to a first detection here
        # and a second one again below

        if os.path.isdir(sim._filename):
            candidates = candidates.union(glob.glob(os.path.join(sim._filename, "*z*particles*")))

        return list(candidates)

    @classmethod
    def _user_specified_filename_to_ahf_basename(cls, filename: str | pathlib.Path) -> pathlib.Path:
        allowed_endings = ["AHF_halos", "AHF_particles", "AHF_", "AHF"]
        filename = util.cutgz(filename)
        for ending in allowed_endings:
            if str(filename).endswith(ending):
                return pathlib.Path(str(filename)[:-len(ending)]+"AHF_")
        raise ValueError("Filename cannot be understood as an AHF catalogue basename. "
                         "For information about how to specify an AHF filename, see the class documentation for "
                         "AHFCatalogue.")

    @classmethod
    def _can_load(cls, sim, filename=None, **kwargs):
        if filename is not None:
            try:
                cls._user_specified_filename_to_ahf_basename(filename)
                return True
            except ValueError:
                return False
        else:
            candidates = cls._list_candidate_ahf_basenames(sim)
            number_ahf_file_candidates = len(candidates)
            return number_ahf_file_candidates > 0



"""Implements classes to automatically copy data from another snapshot when needed.

Most users will not need to use this module directly. It is used by `tangos <https://pynbody.github.io/tangos>`_
to provide a transparent view of a snapshot that is actually stored in a different location.
"""

import copy

from .simsnap import SimSnap


class UnderlyingClassMixin:
    """Mixin for a SimSnap that allows it to derive quantities associated with another class."""
    def __init__(self, underlying_class, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._underlying_class = underlying_class

    def _find_deriving_function(self, name):
        cl = self._underlying_class
        if cl in self._derived_array_registry \
                and name in self._derived_array_registry[cl]:
            return self._derived_array_registry[cl][name]
        else:
            return super()._find_deriving_function(name)

class CopyOnAccessSimSnap(UnderlyingClassMixin, SimSnap):
    """SimSnap that copies data from another SimSnap when that data is needed.

    To the user, data which is already loaded in the underlying snapshot presents merely as 'loadable'
    (i.e. in loadable_keys)."""

    def __init__(self, base: SimSnap, underlying_class=None):
        self._copy_from = base
        if underlying_class is None:
            ancestor = base.ancestor
            if hasattr(ancestor, "_underlying_class"):
                underlying_class = ancestor._underlying_class
            else:
                underlying_class = type(ancestor)

        super().__init__(underlying_class)

        self._unifamily = base._unifamily
        self._file_units_system = base._file_units_system
        self._num_particles = len(base)
        self._family_slice = {f: base._get_family_slice(f) for f in base.families()}
        self._filename = base._filename+":copied_on_access"
        self._dont_try_accessing = []
        self.properties = copy.deepcopy(base.properties)


    def _load_array(self, array_name, fam=None):
        if (array_name, fam) in self._dont_try_accessing:
            raise OSError("Previously tried to get this array without success; not trying again")
        try:
            with self._copy_from.lazy_derive_off, self.lazy_derive_off, self.auto_propagate_off:
                if fam is None:
                    self[array_name] = self._copy_from[array_name]
                else:
                    self[fam][array_name] = self._copy_from[fam][array_name]
        except KeyError as e:
            self._dont_try_accessing.append((array_name, fam))
            raise OSError("Not found in underlying snapshot") from e

    def loadable_keys(self, fam=None):
        if fam is None:
            loaded_keys_in_parent = self._copy_from.keys()
        else:
            loaded_keys_in_parent = self._copy_from.family_keys(fam)
        return list(set(self._copy_from.loadable_keys(fam)).union(loaded_keys_in_parent))



cimport cython
cimport numpy as np

import logging
import math

import numpy as np
from cython.parallel import prange

logger = logging.getLogger('pynbody.analysis._com')


@cython.boundscheck(False)
@cython.wraparound(False)
def shrink_sphere_center(np.ndarray[np.float64_t, ndim=2] pos,
                         np.ndarray[np.float64_t, ndim=1] mass,
                         int min_particles,
                         int particles_for_second_radius,
                         float shrink_factor,
                         float starting_rmax,
                         int num_threads,
                         int itermax=1000) :

    cdef int npart = len(pos)
    cdef int npart_all = len(pos)
    cdef float r2 = 0
    cdef float tot_mass=0
    cdef np.ndarray[np.float64_t, ndim=1] com = np.zeros(3)
    cdef np.ndarray[np.float64_t, ndim=1] com_x = pos.mean(axis=0)
    cdef int i
    cdef int iternum=0
    cdef float current_rmax = np.inf, current_rmax2
    cdef double cx, cy, cz
    cdef double offset_x, offset_y, offset_z
    cdef double pix, piy, piz, mi

    logger.info("Initial rough COM=%s",com_x)

    second_radius = None

    while True :
        offset_x=0; offset_y=0; offset_z=0;
        cx=com_x[0]; cy=com_x[1]; cz=com_x[2]
        current_rmax2 = current_rmax*current_rmax
        with nogil:
            npart = 0
            for i in prange(npart_all, schedule='static', num_threads=num_threads):
                pix=pos[i,0]-cx; piy=pos[i,1]-cy; piz=pos[i,2]-cz
                r2 = pix*pix+piy*piy+piz*piz
                if r2<current_rmax2 :
                    mi = mass[i]
                    offset_x+=pix*mi
                    offset_y+=piy*mi
                    offset_z+=piz*mi
                    tot_mass+=mi
                    npart+=1

        if npart < particles_for_second_radius and second_radius is None:
            second_radius = math.sqrt(current_rmax2)


        if npart<min_particles:
            break


        # divide out total mass and shift
        com[0]=cx+offset_x/tot_mass; com[1]=cy+offset_y/tot_mass; com[2]=cz+offset_z/tot_mass



        # update for next cycle
        com_x[:]= com
        com[:]=0
        tot_mass=0

        iternum+=1
        if iternum>1 :
            current_rmax*=shrink_factor
        else :
            current_rmax = starting_rmax

        if iternum>itermax:
            raise RuntimeError, "shrink_sphere_center failed to converge after %d iterations"%itermax

    return com_x, current_rmax, second_radius



import copy
import gc
from pathlib import Path

import numpy as np
import numpy.testing as npt
import pytest

import pynbody
from pynbody.array import shared


def setup_module():
    global f
    f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")

    # for compatibility with original results, pretend the box
    # is not periodic
    del f.properties['boxsize']

def teardown_module():
    global f
    del f

test_folder = Path(__file__).parent
@pytest.fixture
def v_mean():
    yield np.load(test_folder / 'test_v_mean.npy')

@pytest.fixture
def v_disp():
    yield np.load(test_folder / 'test_v_disp.npy')

@pytest.fixture
def smooth():
    yield np.load(test_folder / 'test_smooth.npy')

@pytest.fixture
def rho():
    yield np.load(test_folder / 'test_rho.npy')

@pytest.fixture
def rho_W():
    yield np.load(test_folder / 'test_rho_W.npy')

@pytest.fixture
def rho_periodic():
    yield np.load(test_folder / 'test_rho_periodic.npy')

@pytest.fixture
def smooth_periodic():
    yield np.load(test_folder / 'test_smooth_periodic.npy')

@pytest.fixture
def div_curl():
    yield np.load(test_folder / 'test_div_curl.npz')

def test_smooth(v_mean, v_disp, rho, smooth):
    global f
    """
    np.save('test_smooth.npy', f.dm['smooth'][::100])
    np.save('test_rho.npy', f.dm['rho'][::100])
    np.save('test_v_mean.npy',f.dm['v_mean'][::100])
    np.save('test_v_disp.npy',f.dm['v_disp'][::100])
    """

    npt.assert_allclose(f.dm['smooth'][::100],
                        smooth,rtol=1e-8)

    npt.assert_allclose(f.dm['rho'][::100],
                        rho,rtol=1e-8)



    npt.assert_allclose(v_mean,f.dm['v_mean'][::100],rtol=1e-8)
    npt.assert_allclose(v_disp,f.dm['v_disp'][::100],rtol=1e-8)

    # check 1D smooth works too
    vz_mean = f.dm.kdtree.sph_mean(f.dm['vz'],32)
    npt.assert_allclose(v_mean[:,2],vz_mean[::100],rtol=1e-8)

    # check 1D dispersions
    v_disp_squared = (
        f.dm.kdtree.sph_dispersion(f.dm['vx'], 32)**2+
        f.dm.kdtree.sph_dispersion(f.dm['vy'], 32)**2+
        f.dm.kdtree.sph_dispersion(f.dm['vz'], 32)**2
    )

    npt.assert_allclose(v_disp**2, v_disp_squared[::100], rtol=1e-8)


def test_smooth_WendlandC2(rho_W):
    pynbody.config['sph']['Kernel'] = 'WendlandC2'

    try:
        f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
        del f.properties['boxsize']
        np.save('test_rho_W2.npy', f.dm['rho'][::100])
        npt.assert_allclose(f.d['rho'][::100], rho_W, rtol=1e-6)
    finally:
        pynbody.config['sph']['Kernel'] = 'CubicSpline'

def test_kd_delete():
    global f
    f.dm['smooth']

    assert hasattr(f.dm,'kdtree')

    f.physical_units()

    # position array has been updated - kdtree should be auto-deleted
    assert not hasattr(f.dm,'kdtree')


def test_kd_issue_88() :
    # number of particles less than number of smoothing neighbours
    f = pynbody.new(gas=16)
    f['pos'] = np.random.uniform(size=(16,3))
    with pytest.raises(ValueError):
        f["smooth"]

@pytest.mark.filterwarnings(r"ignore:overflow.*:RuntimeWarning")
def test_float_kd():
    f = pynbody.load("testdata/gadget2/test_g2_snap")
    del f.properties['boxsize']

    assert f.dm['mass'].dtype==f.dm['pos'].dtype==np.float32
    assert f.dm['smooth'].dtype==np.float32

    # make double copy
    g = pynbody.new(len(f.dm))
    g.dm['pos']=f.dm['pos']
    g.dm['mass']=f.dm['mass']

    assert g.dm['mass'].dtype==g.dm['pos'].dtype==g.dm['smooth'].dtype==np.float64

    # check smoothing lengths agree (they have been calculated differently
    # using floating/double routines)

    npt.assert_allclose(f.dm['smooth'],g.dm['smooth'],rtol=1e-4)
    npt.assert_allclose(f.dm['rho'],g.dm['rho'],rtol=1e-4)

    # check all combinations of float/double smoothing
    double_ar = np.ones(len(f.dm),dtype=np.float64)
    float_ar = np.ones(len(f.dm),dtype=np.float32)

    double_double = g.dm.kdtree.sph_mean(double_ar,32)
    double_float = g.dm.kdtree.sph_mean(float_ar,32)
    float_double = f.dm.kdtree.sph_mean(double_ar,32)
    float_float = f.dm.kdtree.sph_mean(float_ar,32)

    # take double-double as 'gold standard' (though of course if any of these
    # fail it could also be a problem with the double-double case)

    npt.assert_allclose(double_double,double_float,rtol=1e-4)
    npt.assert_allclose(double_double,float_double,rtol=1e-4)
    npt.assert_allclose(double_double,float_float,rtol=1e-4)

def test_periodic_smoothing(rho_periodic, smooth_periodic):
    f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")

    """
    np.save('test_rho_periodic.npy', f.dm['rho'][::100])
    np.save('test_smooth_periodic.npy', f.dm['smooth'][::100])
    """
    npt.assert_allclose(f.dm['rho'][::100],
                         rho_periodic,rtol=1e-5)
    npt.assert_allclose(f.dm['smooth'][::100],
                         smooth_periodic,rtol=1e-5)


@pytest.mark.filterwarnings("ignore:overflow encountered in cast:RuntimeWarning")
def test_neighbour_list():
    f = pynbody.load("testdata/gadget2/test_g2_snap")
    pynbody.sph._get_smooth_array_ensuring_compatibility(f.g)  # actual smoothing
    t = f.g.kdtree
    n_neigh = 32

    generator_nn = t.nn(n_neigh)
    n = next(generator_nn)
    # print(n)

    p_idx = n[0]       # particle index in snapshot arrays
    hsml = n[1]        # smoothing length
    neigh_list = np.array(n[2])  # neighbours list
    dist2 = np.array(n[3])       # squared distances from neighbours
    assert p_idx == 9
    assert hsml == f.g['smooth'][p_idx]
    npt.assert_allclose(hsml,np.sqrt(np.max(dist2))/2, rtol=1e-6)
    assert np.allclose(hsml, 128.19053649902344)

    ordering = np.argsort(dist2)
    assert (neigh_list[ordering] == [   9,   37,    8,   10,   24,   35, 2018, 2017,   25,   41, 1998,
         20,   40,   34,   38,   11,   52, 1997,   36,   22,   19,   43,
         21,   12,   42,    7,    5, 1996,   33,   39,   23,   31]).all()
    npt.assert_allclose(dist2[ordering], [    0.        , 16521.75195312, 16879.42578125, 17574.50195312,
       20319.2890625 , 24460.67773438, 24489.19140625, 29066.84765625,
       30422.66796875, 31192.06835938, 31658.59375   , 34860.97265625,
       35900.76953125, 36082.15234375, 36883.796875  , 39369.51953125,
       41815.23046875, 45676.2578125 , 47861.25390625, 51718.3515625 ,
       52811.79296875, 54654.58984375, 56239.8828125 , 57026.3984375 ,
       58157.92578125, 58536.9765625 , 59311.27734375, 59870.70703125,
       60277.04296875, 60706.04296875, 61944.99609375, 65731.2578125 ],
                        rtol=1e-6)

    neighbour_list_all = t.all_nn(n_neigh)
    assert n == neighbour_list_all[0]
    for nl in neighbour_list_all:
        assert len(nl[2]) == n_neigh   # always find n_neigh neighbours
        idx_self = nl[2].index(nl[0])  # index of self in the neighbour list (not necessarily the first element)
        assert nl[3][idx_self] == 0.0  # distance to self

def test_div_curl_smoothing(div_curl):
    f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")

    """
    np.savez('test_div_curl', curl=f.g['v_curl'][::100], div=f.g['v_div'][::100])
    """
    arr = div_curl
    # print(f.g['v_curl'][::100], f.g['v_div'][::100])
    curl, div = arr['curl'], arr['div']
    npt.assert_allclose(f.g['v_curl'][::100], curl, rtol=2e-4)
    npt.assert_allclose(f.g['v_div'][::100],  div,  rtol=2e-4)
    npt.assert_equal(f.g['vorticity'], f.g['v_curl'])
    assert f.g['vorticity'].units == f.g['vel'].units/f.g['pos'].units

def test_kdtree_parallel_build():
    """Check that parallel tree build results in identical tree to serial build."""
    f = pynbody.new(dm=5000)
    f['pos'] = np.random.uniform(size=(5000,3))
    f['mass'] = np.random.uniform(size=5000)

    f.build_tree(1)
    result_one_thread = f.kdtree.serialize()

    _, _, kdn1, poff1, _ = result_one_thread

    del f.kdtree

    f.build_tree(4)
    result_four_threads = f.kdtree.serialize()
    _, _, kdn4, poff4, _ = result_four_threads

    assert (kdn1['pLower'] == kdn4['pLower']).all()
    assert (kdn1['pUpper'] == kdn4['pUpper']).all()
    assert (kdn1['iDim'] == kdn4['iDim']).all()
    npt.assert_allclose(kdn1['bnd']['fMin'], kdn4['bnd']['fMin'])
    npt.assert_allclose(kdn1['bnd']['fMax'], kdn4['bnd']['fMax'])
    npt.assert_allclose(kdn1['fSplit'], kdn4['fSplit'])
    assert (poff1 == poff4).all()






@pytest.mark.parametrize("npart", [1, 10, 100, 1000, 100000])
@pytest.mark.parametrize("offset", [0.0, 0.2, 0.5]) # checks wrapping
@pytest.mark.parametrize("radius", [0.1, 0.3, 1.0])
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_particles_in_sphere(npart, offset, radius, dtype):
    f = pynbody.new(dm=npart)

    f._create_array('pos', 3, dtype)
    f._create_array('mass', 1, dtype)

    np.random.seed(1337)
    f['pos'] = np.random.uniform(low=-0.5, high=0.5, size=(npart,3))
    f['mass'] = np.random.uniform(size=npart)
    assert np.issubdtype(f['pos'].dtype, dtype)
    assert np.issubdtype(f['mass'].dtype, dtype)
    f.properties['boxsize'] = 1.0

    f.build_tree()
    particles = f.kdtree.particles_in_sphere([offset, 0.0, 0.0], radius)

    f['x'] -= offset
    f.wrap()
    particles_compare = np.where(f['r']<radius)[0]

    assert (np.sort(particles) == np.sort(particles_compare)).all()

def test_kdtree_from_existing_kdtree(npart=1000):
    f = _make_test_gaussian(npart)

    f_copy = copy.deepcopy(f)

    f.build_tree()
    f_copy.import_tree(f.kdtree.serialize())

    assert f_copy.kdtree is not f.kdtree

    npt.assert_allclose(f['smooth'], f_copy['smooth'], atol=1e-7)


def _make_test_gaussian(npart):
    f = pynbody.new(dm=npart)
    np.random.seed(1337)
    f['pos'] = np.random.normal(1.0, size=(npart, 3))
    f['mass'] = np.random.uniform(size=npart)
    return f


def test_kdtree_shared_mem(npart=1000):
    f = _make_test_gaussian(npart)
    n = shared.get_num_shared_arrays_owned()
    f.build_tree(shared_mem=False)
    assert shared.get_num_shared_arrays_owned() == n
    del f

    f = _make_test_gaussian(npart)
    f.build_tree(shared_mem=True)
    assert shared.get_num_shared_arrays_owned() == 2 + n
    assert f.kdtree.kdnodes._shared_fname.startswith('pynbody')
    assert f.kdtree.particle_offsets._shared_fname.startswith('pynbody')
    del f
    gc.collect()
    assert shared.get_num_shared_arrays_owned() == n

def test_boxsize_too_small():
    f = pynbody.new(dm=1000)
    f['pos'] = np.random.normal(scale=1.0, size=f['pos'].shape)
    f['vel'] = np.random.normal(scale=1.0, size=f['vel'].shape)
    f['mass'] = np.random.uniform(1.0, 10.0, size=f['mass'].shape)
    f.properties['boxsize'] = 0.1
    with pytest.warns(RuntimeWarning, match = "span a region larger than the specified boxsize"):
        _ = f['smooth']



"""Module for describing, applying and reverting Galilean transformations on simulations.

Generally it is not necessary to access this module directly, but it is used by
:class:`~pynbody.snapshot.simsnap.SimSnap` class to implement e.g.
:meth:`~pynbody.snapshot.simsnap.SimSnap.rotate_x` etc, and also by analysis
transformations such as :func:`~pynbody.analysis.halo.center` and
:func:`~pynbody.analysis.angmom.faceon` etc.

The main feature of this module is the :class:`Transformation` class, which performs
transformations of various kinds on simulations then reverts them.  You can use these objects
as context managers, which means you can write code like this:

>>> with f.translate(fshift) :
>>>     print(f['pos'][0])

which is equivalent to

>>> try:
>>>     f['pos']+=shift
>>>     print(f['pos'][0])
>>> finally:
>>>     f['pos']-=shift

You can also chain transformations together, like this:

>>> with f.translate(fshift).rotate_x(45):
>>>    print(f['pos'][0])

This implies a translation followed by a rotation. When reverting the transformation, they are
of course undone in the opposite order.
"""

from __future__ import annotations

import typing
import weakref

if typing.TYPE_CHECKING:
    from . import snapshot

import numpy as np

from . import util


class Transformable:
    """A mixin class for objects that can generate a Transformation object"""

    def translate(self, offset):
        """Translate by the given offset.

        Returns a :class:`pynbody.transformation.GenericTranslation` object which can be used
        as a context manager to ensure that the translation is undone.

        For more information, see the :mod:`pynbody.transformation` documentation."""
        return GenericTranslation(self, 'pos', offset, description="translate")

    def offset_velocity(self, offset):
        """Shift the velocity by the given offset.

        Returns a :class:`pynbody.transformation.GenericTranslation` object which can be used
        as a context manager to ensure that the translation is undone.

        For more information, see the :mod:`pynbody.transformation` documentation."""
        return GenericTranslation(self, 'vel', offset, description = "offset_velocity")

    def rotate_x(self, angle):
        """Rotates about the current x-axis by 'angle' degrees.

        Returns a :class:`pynbody.transformation.GenericTranslation` object which can be used
        as a context manager to ensure that the translation is undone.

        For more information, see the :mod:`pynbody.transformation` documentation."""
        angle_rad = angle * np.pi / 180
        return self.rotate(np.array([[1, 0, 0],
                                     [0, np.cos(angle_rad), -np.sin(angle_rad)],
                                     [0, np.sin(angle_rad),  np.cos(angle_rad)]]),
                           description = f"rotate_x({angle})")

    def rotate_y(self, angle):
        """Rotates about the current y-axis by 'angle' degrees.

        Returns a :class:`pynbody.transformation.GenericTranslation` object which can be used
        as a context manager to ensure that the translation is undone.

        For more information, see the :mod:`pynbody.transformation` documentation."""
        angle_rad = angle * np.pi / 180
        return self.rotate(np.array([[np.cos(angle_rad), 0, np.sin(angle_rad)],
                                     [0,                1,        0],
                                     [-np.sin(angle_rad),   0,   np.cos(angle_rad)]]),
                           description = f"rotate_y({angle})")

    def rotate_z(self, angle):
        """Rotates about the current z-axis by 'angle' degrees.

        Returns a :class:`pynbody.transformation.GenericTranslation` object which can be used
        as a context manager to ensure that the translation is undone.

        For more information, see the :mod:`pynbody.transformation` documentation."""
        angle_rad = angle * np.pi / 180
        return self.rotate(np.array([[np.cos(angle_rad), -np.sin(angle_rad), 0],
                                     [np.sin(angle_rad),  np.cos(angle_rad), 0],
                                     [0,             0,        1]]),
                           description = f"rotate_z({angle})")

    def rotate(self, matrix, description = None):
        """Rotates using a specified matrix.

        Returns a :class:`pynbody.transformation.GenericTranslation` object which can be used
        as a context manager to ensure that the translation is undone.

        For more information, see the :mod:`pynbody.transformation` documentation.

        Parameters
        ----------

        matrix : array_like
            The 3x3 orthogonal matrix to rotate by

        description : str
            A description of the rotation to be returned from str() and repr()
        """
        return Rotation(self, matrix, description = description)

    @util.deprecated("This method is deprecated and will be removed in a future version. Use the rotate method instead.")
    def transform(self, matrix):
        """Deprecated alias for :meth:`rotate`."""
        return self.rotate(matrix)


class Transformation(Transformable):
    """The base class for all transformations.

    Note that this class inherits from :class:`Transformable`, so all transformations are themselves
    transformable. This means that you can chain transformations together, e.g. for a
    :class:`~pynbody.snapshot.simsnap.SimSnap` object *f*:

    >>> with f.translate(fshift).rotate_x(45):
    >>>    ...
    """

    def __init__(self, f, defer = False, description = None):
        """Initialise a transformation, and apply it if not explicitly deferred

        Parameters
        ----------
        f : SimSnap or Transformation
            The simulation or transformation to act on. If a transformation is given, this
            transformation will be chained to it, i.e. the result will represent the composed
            transformation.
        defer : bool
            If True, the transformation is not applied immediately. Otherwise, as soon as the object
            is constructed the transformation is applied to the simulation
        description : str
            A description of the transformation to be returned from str() and repr()
        """
        from . import snapshot

        if isinstance(f, NullTransformation):
            f = f.sim # as though we are starting from the simulation itself

        if isinstance(f, snapshot.SimSnap):
            self.sim = f
            self.next_transformation = None
        elif isinstance(f, Transformation):
            self.sim = f.sim
            self.next_transformation = f
        else:
            raise TypeError("Transformation must either act on another Transformation or on a SimSnap")

        self._description = description

        self.applied = False
        if not defer:
            self.apply(force=False)

    @property
    def sim(self) -> snapshot.SimSnap | None:
        """The simulation to which this transformation applies"""
        return self._sim()

    @sim.setter
    def sim(self, sim):
        if sim is None:
            self._sim = lambda: None
        else:
            self._sim = weakref.ref(sim)

    def __repr__(self):
        return "<Transformation " + str(self) + ">"

    def __str__(self):
        if self.next_transformation is not None:
            s = str(self.next_transformation)+", "
        else:
            s = ""

        return s + self._describe()

    def _describe(self):
        if self._description:
            return self._description
        else:
            return self.__class__.__name__

    def apply_to(self, f: snapshot.SimSnap) -> snapshot.SimSnap:
        """Apply this transformation to a specified simulation.

        Chained transformations are applied recursively.

        Parameters
        ----------
        f : SimSnap
            The simulation to apply the transformation to. Any simulation reference stored within
            the transformation itself is ignored

        Returns
        -------
        SimSnap
            The input simulation (not a copy)
        """
        if self.next_transformation is not None:
            # this is a chained transformation, get the SimSnap to operate on
            # from the level below
            f = self.next_transformation.apply_to(f)

        self._apply(f)

        return f

    def apply_inverse_to(self, f):
        """Apply the inverse of this transformation to a specified simulation.

        Chained transformations are applied recursively.

        Parameters
        ----------
        f : SimSnap
            The simulation to apply the transformation to

        Returns
        -------
        SimSnap
            The input simulation (not a copy)
        """
        self._revert(f)

        if self.next_transformation is not None:
            self.next_transformation.apply_inverse_to(f)



    def apply(self, force=True):
        """Apply the transformation to the simulation it is associated with.

        This is either the simulation passed to the constructor or the one passed to the last
        transformation in the chain. If the transformation has already been applied, a RuntimeError
        is raised unless *force* is True, in which case the transformation is applied again.

        Parameters
        ----------

        force : bool
            If True, the transformation is applied even if it has already been applied. Otherwise,
            a RuntimeError is raised if the transformation has already been applied.

        Returns
        -------

        SimSnap
            The simulation after the transformation has been applied
        """

        if self.next_transformation is not None:
            # this is a chained transformation, get the SimSnap to operate on
            # from the level below
            f = self.next_transformation.apply(force=force)
        else:
            f = self.sim

        if self.applied and force:
            raise RuntimeError("Transformation has already been applied")

        if not self.applied:
            self._apply(f)
            self.applied = True
            self.sim = f

        return f

    def revert(self):
        """Revert the transformation. If it has not been applied, a RuntimeError is raised."""
        if not self.applied:
            raise RuntimeError("Transformation has not been applied")
        self._revert(self.sim)
        self.applied = False
        if self.next_transformation is not None:
            self.next_transformation.revert()

    def _apply(self, f):
        pass

    def _revert(self, f):
        pass

    def __enter__(self):
        self.apply(force=False)
        return self

    def __exit__(self, *args):
        self.revert()


class NullTransformation(Transformation):
    """A transformation that does nothing, provided for convenience where a transformation is expected."""

    def __init__(self, f):
        super().__init__(f, description="null")

    def _apply(self, f):
        pass
    def _revert(self, f):
        pass


class GenericTranslation(Transformation):
    """A translation on a specified array of a simulation"""

    def __init__(self, f, arname, shift, description = None):
        """Initialise a translation on a named array

        Parameters
        ----------
        f : Transformable
            The transformable to act on
        arname : str
            The name of the array to translate
        shift : array_like
            The shift to apply
        description : str
            A description of the translation to be returned from str() and repr()
        """
        self.shift = shift
        self.arname = arname
        super().__init__(f, description=description)

    def _apply(self, f):
        f[self.arname] += self.shift

    def _revert(self, f):
        f[self.arname] -= self.shift


class Rotation(Transformation):
    """A rotation on all 3d vectors in a simulation, by a given orthogonal 3x3 matrix"""

    def __init__(self, f, matrix, ortho_tol=1.e-8, description = None):
        """Initialise a rotation on a simulation.

        The matrix must be orthogonal to within *ortho_tol*.

        Parameters
        ----------
        f : SimSnap
            The simulation to act on

        matrix : array_like
            The 3x3 orthogonal matrix to rotate by

        ortho_tol : float
            The tolerance for orthogonality of the matrix. If the matrix is not orthogonal to within
            this tolerance, a ValueError is raised.

        description: str
            A description of the rotation to be returned from str() and repr()


        """
        # Check that the matrix is orthogonal
        resid = np.dot(matrix, np.asarray(matrix).T) - np.eye(3)
        resid = (resid ** 2).sum()
        if resid > ortho_tol or resid != resid:
            raise ValueError("Transformation matrix is not orthogonal")
        self.matrix = matrix
        if description is None:
            description = "rotate"
        super().__init__(f, description=description)

    def _apply(self, f):
        self._transform(self.matrix)

    def _revert(self, f):
        self._transform(self.matrix.T)

    def _transform(self, matrix):
        """Transforms the snapshot according to the 3x3 matrix given."""

        sim = self.sim

        # NB though it might seem more efficient to access _arrays and
        # _family_arrays directly, this would not work for SubSnaps.
        snapshot_keys = sim.keys()

        for array_name in snapshot_keys:
            ar = sim[array_name]
            if (not ar.derived) and len(ar.shape) == 2 and ar.shape[1] == 3:
                ar[:] = np.dot(matrix, ar.transpose()).transpose()

        for fam in sim.families():
            family_keys = sim[fam].keys()
            family_keys_not_in_snapshot = set(family_keys) - set(snapshot_keys)
            for array_name in family_keys_not_in_snapshot:
                ar = sim[fam][array_name]
                if (not ar.derived) and len(ar.shape) == 2 and ar.shape[1] == 3:
                    ar[:] = np.dot(matrix, ar.transpose()).transpose()


GenericRotation = Rotation # name from pynbody v1


@util.deprecated("This function is deprecated and will be removed in a future version. Use the translate method of a SimSnap object instead.")
def translate(f, shift):
    """Deprecated alias for ``f.translate(shift)``"""

    return GenericTranslation(f, 'pos', shift)

@util.deprecated("This function is deprecated and will be removed in a future version. Use the translate method of a SimSnap object instead.")
def inverse_translate(f, shift):
    """Deprecated alias for ``f.translate(-shift)``"""
    return translate(f, -np.asarray(shift))

@util.deprecated("This function is deprecated and will be removed in a future version. Use the offset_velocity method of a SimSnap object instead.")
def v_translate(f, shift):
    """Deprecated alias for ``f.offset_velocity(shift)``"""

    return GenericTranslation(f, 'vel', shift)

@util.deprecated("This function is deprecated and will be removed in a future version. Use the offset_velocity method of a SimSnap object instead.")
def inverse_v_translate(f, shift):
    """Deprecated alias for ``f.offset_velocity(-shift)``"""

    return GenericTranslation(f, 'vel', -np.asarray(shift))

@util.deprecated("This function is deprecated and will be removed in a future version. Use sim.translate(...).vel_translate(...) instead.")
def xv_translate(f, x_shift, v_shift):
    """Deprecated alias for ``f.translate(x_shift).offset_velocity(v_shift)``"""

    return translate(v_translate(f, v_shift), x_shift)

@util.deprecated("This function is deprecated and will be removed in a future version. "
                 "Use sim.translate(...).vel_translate(...) instead.")
def inverse_xv_translate(f, x_shift, v_shift):
    """Deprecated alias for ``f.translate(-x_shift).offset_velocity(-v_shift)``"""

    return translate(v_translate(f, -np.asarray(v_shift)),
                     -np.asarray(x_shift))

@util.deprecated("This function is deprecated and will be removed in a future version. "
                 "Use the rotate method of a SimSnap object instead.")
def transform(f, matrix):
    """Deprecated alias for ``f.rotate(matrix)``"""

    return Rotation(f, matrix)


@util.deprecated("This function is deprecated and will be removed in a future version. Use NullTransformation instead.")
def null(f):
    """Deprecated alias for ``NullTransformation(f)``"""

    return NullTransformation(f)



"""

hifrac
=======

calculates Hydrogen ionization fraction - limited version of ionfrac to make use of CLOUDY HI table

"""

import logging
import os

import numpy as np

from ..array import SimArray
from .interpolate import interpolate3d

logger = logging.getLogger('pynbody.analysis.hifrac')

from pynbody import config


def calculate(sim,ion='hi',selfshield=False) :
    """

    calculate -- documentation placeholder

    """
    import h5py
    global config
    # ionization fractions calculated for optically thin case with
    # CLOUDY for Duffy et al. (2012) overall h1frac.py based on
    # ionfrac.py routine, ultimately should be merged when other
    # elements are added to the hdf5 file
    iffile = os.path.join(os.path.dirname(__file__),"h1.hdf5")
    if os.path.exists(iffile) :
        # import data
        logger.info("Loading %s" % iffile)
        ifs=h5py.File(iffile,'r')
    else :
        raise OSError("h1.hdf5 (HI Fraction table) not found")

    # allocate temporary metals that we can play with
    # before inlining, the views on the arrays must be standard np.ndarray
    # otherwise the normal numpy macros are not generated
    x_vals = ifs['logd'][:].view(np.ndarray)
    y_vals = ifs['logt'][:].view(np.ndarray)
    z_vals = ifs['redshift'][:].view(np.ndarray)
    vals = np.log10(ifs['ionbal'][:]).view(np.ndarray)
    ifs.close()

    z = np.zeros(len(sim.gas)).view(np.ndarray)
    z[:] = sim.properties['z']
    y = np.log10(sim.gas['temp']).view(np.ndarray)
    x = np.log10(sim.gas['rho'].in_units('m_p cm^-3')).view(np.ndarray)
    n = len(sim.gas)
    n_x_vals = len(x_vals)
    n_y_vals = len(y_vals)
    n_z_vals = len(z_vals)
    result_array = np.zeros(n)

    # get values off grid to minmax
    x[np.where(x < np.min(x_vals))] = np.min(x_vals)
    x[np.where(x > np.max(x_vals))] = np.max(x_vals)
    y[np.where(y < np.min(y_vals))] = np.min(y_vals)
    y[np.where(y > np.max(y_vals))] = np.max(y_vals)
    z[np.where(z < np.min(z_vals))] = np.min(z_vals)
    z[np.where(z > np.max(z_vals))] = np.max(z_vals)

    #interpolate
    logger.info("Interpolation %s values" % ion)
    result_array = interpolate3d(x, y, z, x_vals, y_vals, z_vals, vals)

    ## Selfshield criteria assume all EoS gas
    if selfshield != False:
        result_array[sim.gas['OnEquationOfState'] == 1.] = 0.
    ## Selfshield criteria from Duffy et al 2012a (in addition to EoS gas)
        if selfshield == 'duffy12':
            result_array[(sim.gas['p'].in_units('m_p K cm**-3') > 150.) & (sim.gas['temp'].in_units('K') < 10.**(4.5))] = 0.

    ## Get as HI per proton mass (essentially multiplying the HI fraction by the Hydrogen mass fraction)
    result_array += np.log10(sim.gas['hydrogen'])

    result_array = (10.**result_array).view(SimArray)
#    result_array.units = units.m_p**-1

    return result_array



"""Support for generic halo catalogues based on an array of halo numbers for each particle."""

import numpy as np

from . import HaloCatalogue
from .details.number_mapping import create_halo_number_mapper
from .details.particle_indices import HaloParticleIndices


class HaloNumberCatalogue(HaloCatalogue):
    """A generic catalogue using an array of halo numbers, one per particle.

    This is the output format used by SKID, for example.
    """
    def __init__(self, sim, array='grp', ignore=None, **kwargs):
        """Construct a GrpCatalogue, extracting halos based on a simulation-wide integer array with their numbers

        *sim* - the SimSnap for which the halos will be constructed
        *array* - the name of the array which should be present, loadable or derivable across the simulation
        *ignore* - a special value indicating "no halo", or None if no such special value is defined
        """
        sim[array] # noqa - trigger lazy-loading and/or kick up a fuss if unavailable
        self._array = array
        self._ignore = ignore
        self._halo_numbers = np.unique(sim[array])
        number_mapper = create_halo_number_mapper(self._trim_array_for_ignore(self._halo_numbers))
        HaloCatalogue.__init__(self, sim, number_mapper=number_mapper)

    def _trim_array_for_ignore(self, array):
        assert len(array) == len(self._halo_numbers)
        if self._ignore is None:
            return array
        if self._ignore == self._halo_numbers[0]:
            return array[1:]
        elif self._ignore == self._halo_numbers[-1]:
            return array[:-1]
        else:
            raise ValueError(
                "ignore must be either the smallest or largest value in the array")


    def _get_all_particle_indices(self):

        halo_number_per_particle = self.base[self._array]

        particle_index_list = np.argsort(halo_number_per_particle, kind='mergesort')
        start = np.searchsorted(halo_number_per_particle[particle_index_list], self._halo_numbers)
        stop = np.concatenate((start[1:], [len(particle_index_list)]))

        particle_index_list_boundaries = self._trim_array_for_ignore(
            np.hstack((start[:, np.newaxis], stop[:, np.newaxis]))
        )

        return HaloParticleIndices(particle_ids = particle_index_list, boundaries = particle_index_list_boundaries)

    def _get_particle_indices_one_halo(self, halo_number):
        if halo_number == self._ignore:
            self._no_such_halo(halo_number)
        array = np.where(self.base[self._array] == halo_number)[0]
        if len(array) == 0:
            self._no_such_halo(halo_number)
        return array

    def _no_such_halo(self, i):
        raise KeyError(f"No such halo {i}")

    def get_group_array(self, family=None, use_index=False):
        if family is not None:
            result = self.base[family][self._array]
        else:
            result = self.base[self._array]

        if use_index:
            return self.number_mapper.number_to_index(result)
        else:
            return result


    @classmethod
    def _can_load(cls, sim, arr_name='grp'):
        if (arr_name in sim.loadable_keys()) or (arr_name in list(sim.keys())) :
            return True
        else:
            return False

class AmigaGrpCatalogue(HaloNumberCatalogue):
    """A catalogue of halos using Alyson Brooks' post-processed AHF output (turned into a SKID-like array)"""
    def __init__(self, sim):
        super().__init__(sim, array='amiga.grp')

    @classmethod
    def _can_load(cls, sim, arr_name='amiga.grp'):
        return super()._can_load(sim, arr_name)



"""
Functions for common cosmological calculations.

"""

import math

import numpy as np
import scipy
import scipy.integrate
from scipy.interpolate import interp1d

from .. import units
from ..array import SimArray
from ..configuration import config_parser

_interp_points = int(config_parser.get('general','cosmo-interpolation-points'))


def _a_dot(a, h0, om_m, om_l):
    om_k = 1.0 - om_m - om_l
    return h0 * a * np.sqrt(om_m * (a ** -3) + om_k * (a ** -2) + om_l)


def _a_dot_recip(*args):
    return 1. / _a_dot(*args)


def _da_dtau(a, h0, om_m, om_l):
    return a**2 * _a_dot(a, h0, om_m, om_l)

def _da_dtau_recip(*args):
    return 1. / _da_dtau(*args)


def _hzoverh0(a, omegam0):
    """ returns: H(a) / H0  = [omegam/a**3 + (1-omegam)]**0.5 """
    return np.sqrt(omegam0 * np.power(a, -3) + (1. - omegam0))


def _lingrowthintegrand(a, omegam0):
    """ (e.g. eq. 8 in lukic et al. 2008)   returns: da / [a*H(a)/H0]**3 """
    return np.power((a * _hzoverh0(a, omegam0)), -3)


def _lingrowthfac(red, omegam0, omegal0, return_norm=False):
    """
    returns: linear growth factor, b(a) normalized to 1 at z=0, good for flat lambda only
    a = 1/1+z
    b(a) = Delta(a) / Delta(a=1)   [ so that b(z=0) = 1 ]
    (and b(a) [Einstein de Sitter, omegam=1] = a)

    Delta(a) = 5 omegam / 2 H(a) / H(0) * integral[0:a] [da / [a H(a) H0]**3]
    equation  from  peebles 1980 (or e.g. eq. 8 in lukic et al. 2008) """

    # need to add w ~= , nonflat, -1 functionality

    import scipy.integrate

    if (abs(omegam0 + omegal0 - 1.) > 1.e-4):
        raise RuntimeError("Linear growth factors can only be calculated for flat cosmologies")

    a = 1 / (1. + red)

    # 1st calc. for z=z
    lingrowth = scipy.integrate.quad(_lingrowthintegrand, 0., a, (omegam0))[0]
    lingrowth *= 5. / 2. * omegam0 * _hzoverh0(a, omegam0)

    # then calc. for z=0 (for normalization)
    a0 = 1.
    lingrowtha0 = scipy.integrate.quad(
        _lingrowthintegrand, 0., a0, (omegam0))[0]
    lingrowtha0 *= 5. / 2. * omegam0 * _hzoverh0(a0, omegam0)

    lingrowthfactor = lingrowth / lingrowtha0
    if return_norm:
        return lingrowthfactor, lingrowtha0
    else:
        return lingrowthfactor


def linear_growth_factor(f, z=None):
    """Calculate the linear growth factor b(a), normalized to 1 at z=0, for the cosmology of snapshot f.

    The output is dimensionless. If a redshift z is specified, it is used in place of the redshift in
    output f.
    """
    if z is None:
        z = f.properties['z']
    omegam0 = f.properties['omegaM0']
    omegal0 = f.properties['omegaL0']
    return _lingrowthfac(z, omegam0, omegal0)


def rate_linear_growth(f, z=None, unit='h Gyr^-1'):
    """Calculate the linear growth rate b'(a), normalized to 1 at z=0, for the cosmology of snapshot f.

    The output is in 'h Gyr^-1' by default, but other units can be specified. If a redshift z is specified,
    it is used in place of the redshift in output f.

    """

    if z is None:
        z = f.properties['z']
    a = 1. / (1. + z)
    omegam0 = f.properties['omegaM0']
    omegal0 = f.properties['omegaL0']

    b, X = _lingrowthfac(z, omegam0, omegal0, return_norm=True)
    I = _lingrowthintegrand(a, omegam0)

    term1 = -(1.5 * omegam0 * a ** -3) * b / \
        math.sqrt(1. - omegam0 + omegam0 * a ** -3)
    term2 = (2.5 * omegam0) * _hzoverh0(a, omegam0) ** 2 * a * I / X

    res = units.h * (term1 + term2) * 100. * units.Unit("km s^-1 Mpc^-1")

    return res.in_units(unit, **f.conversion_context())


def _test_rate_linear_growth(f, z=None, unit='h Gyr^-1'):
    # coded up by AP to test linear growth *rate* equation above
    if z is None:
        z = f.properties['z']
    a0 = 1. / (1. + z)
    a1 = a0 * 0.999
    z0 = 1. / a0 - 1
    z1 = 1. / a1 - 1

    b0 = linear_growth_factor(f, z0)
    b1 = linear_growth_factor(f, z1)

    db = b1 - b0

    unit = units.Unit(unit)
    dt = age(f, z1, unit ** -1) - age(f, z0, unit ** -1)

    return db / dt


def age(f, z=None, unit='Gyr'):
    """
    Calculate the age of the universe in the snapshot f by integrating the Friedmann equation.

    The output is given in the specified units. If a redshift z is specified, it is used in place of the redshift in the
    output f.

    If a long array of redshifts is provided, interpolation is used to speed up the calculation. Specifically,
    the number of interpolation points is controlled by the configuration parameter 'cosmo-interpolation-points'.

    Parameters
    ----------

    f : SimSnap
        The snapshot from which to obtain the cosmological parameters.

    z : float, list, or np.ndarray, optional
        The redshift(s) at which to calculate the age of the universe. If None, the redshift of the snapshot is used.

    unit : str, optional
        The units in which to return the age of the universe. Default is 'Gyr'.

    Returns
    -------

    SimArray | float
        The age of the universe at the specified redshift(s).

    """
    if z is None:
        z = f.properties['z']

    h0 = f.properties['h']
    omM = f.properties['omegaM0']
    omL = f.properties['omegaL0']

    conv = units.Unit("0.01 s Mpc km^-1").ratio(unit, **f.conversion_context())

    def get_age(x):
        x = 1.0 / (1.0 + x)
        return scipy.integrate.quad(_a_dot_recip, 0, x, (h0, omM, omL))[0] * conv

    if isinstance(z, np.ndarray) or isinstance(z, list):
        if len(z) > _interp_points:
            a_vals = np.logspace(-3,0, _interp_points)
            z_vals = 1./a_vals-1.
            log_age_vals = np.log(age(f, z_vals))
            interp = interp1d(np.log(a_vals), log_age_vals, bounds_error=False)
            log_a_input = np.log(1./(1.+z))
            results = np.exp(interp(log_a_input))
        else:
            results = np.array([get_age(_z) for _z in z])
        results = results.view(SimArray)
        results.units = unit
        return results
    else:
        return get_age(z)

def tau(f, z=None, unit="Gyr"):
    """
    Calculate the conformal age of the universe in the snapshot f by integrating the conformal Friedmann equation.

    The output is given in the specified units. If a redshift z is specified, it is used in place of the redshift in the
    output f.

    Parameters
    ----------

    f : SimSnap
        The snapshot from which to obtain the cosmological parameters.

    z : float, list, or np.ndarray, optional
        The redshift(s) at which to calculate the age of the universe. If None, the redshift of the snapshot is used.

    unit : str, optional
        The units in which to return the age of the universe. Default is 'Gyr'.

    Returns
    -------

    SimArray | float
        The conformal age of the universe at the specified redshift(s).


    """
    if z is None:
        z = f.properties['z']

    h0 = f.properties['h']
    omM = f.properties['omegaM0']
    omL = f.properties['omegaL0']

    conv = units.Unit("0.01 s Mpc km^-1").ratio(unit, **f.conversion_context())

    @np.vectorize
    def get_tau(z):
        aexp = 1.0 / (1.0 + z)
        return scipy.integrate.quad(
            _da_dtau_recip,
            1,
            aexp,
            args=(h0, omM, omL)
        )[0]

    results = (get_tau(z) * conv)

    if isinstance(results, np.ndarray):
        results = results.view(SimArray)
        results.units = unit

    return results


@units.takes_arg_in_units((1, "Gyr"), context_arg=0)
def redshift(f, time):
    """
    Calculate the redshift given a snapshot and a time since Big Bang in Gyr.

    Uses scipy.optimize.newton to do the root finding if number of
    elements in the time array is less than 1000; otherwise uses a linear
    interpolation.


    Parameters
    ----------

    f : SimSnap
        The snapshot from which to obtain the cosmological parameters.

    time : float, list, or np.ndarray
        The time(s) since the Big Bang at which to calculate the redshift.

    Returns
    -------
    float or np.ndarray
        The redshift(s) corresponding to the input time(s).

    """

    from scipy.interpolate import interp1d
    from scipy.optimize import newton

    def func(x, sim, time):
        return age(sim, x) - time

    if isinstance(time, list) or isinstance(time, np.ndarray):
        if len(time) > _interp_points:
            zs = np.logspace(3, -10, _interp_points)
            ages = age(f, zs)
            i = interp1d(ages, zs)
            return i(time)
        else:
            return np.array([newton(func, 1, args=(f, x)) for x in time])
    else:
        return newton(func, 1, args=(f, time))


def rho_crit(f, z=None, unit=None):
    """Calculate the critical density of the universe in the snapshot f.

    .. warning ::
        You can get slightly confusing results if your simulation is in comoving units and you specify a different
        redshift z. Specifically, the physical density for the redshift you specify is calulated, but expressed as
        a comoving density *at the redshift of the snapshot*. This is intentional, but can be surprising.

    Parameters
    ----------

    f : SimSnap
        The snapshot from which to obtain the cosmological parameters.

    z : float, optional
        The redshift at which to calculate the critical density. If None, the redshift of the snapshot is used.

    unit : str, optional
        The units in which to return the critical density. If None, the returned density will be in the units of
        f["mass"].units/f["pos"].units**3. If that unit cannot be calculated, the returned units are Msol kpc^-3
        comoving.

    Returns
    -------

    float
        The critical density of the universe at the specified redshift.



    """

    if z is None:
        z = f.properties['z']

    if unit is None:
        try:
            unit = f.dm["mass"].units / f.dm["pos"].units ** 3
        except units.UnitsException:
            unit = units.NoUnit()

    if hasattr(unit, "_no_unit"):
        unit = units.Unit("Msol kpc^-3 a^-3")

    omM = f.properties['omegaM0']
    omL = f.properties['omegaL0']
    h0 = f.properties['h']
    a = 1.0 / (1.0 + z)

    H_z = _a_dot(a, h0, omM, omL) / a
    H_z = units.Unit("100 km s^-1 Mpc^-1") * H_z

    rho_crit = (3 * H_z ** 2) / (8 * math.pi * units.G)

    return rho_crit.ratio(unit, **f.conversion_context())


def rho_M(f, z=None, unit=None):
    """Calculate the matter density of the universe in snapshot f.

    .. warning ::
        You can get slightly confusing results if your simulation is in comoving units and you specify a different
        redshift z. Specifically, the physical density for the redshift you specify is calulated, but expressed as
        a comoving density *at the redshift of the snapshot*. This is intentional, but can be surprising.

    Parameters
    ----------

    f : SimSnap
        The snapshot from which to obtain the cosmological parameters.

    z : float, optional
        The redshift at which to calculate the critical density. If None, the redshift of the snapshot is used.

    unit : str, optional
        The units in which to return the matter density. If None, the returned density will be in the units of
        f["mass"].units/f["pos"].units**3. If that unit cannot be calculated, the returned units are Msol kpc^-3
        comoving.

    Returns
    -------

    float
        The matter density of the universe at the specified redshift.

    """

    if z is None:
        z = f.properties['z']

    return f.properties['omegaM0'] * rho_crit(f, 0, unit) * (1.0 + z) ** 3


def H(f):
    """Calculate the Hubble parameter of the universe in snapshot f"""
    return f.properties['h'] * _hzoverh0(f.properties['a'], f.properties['omegaM0']) * units.Unit("100 km s^-1 Mpc^-1")


def add_hubble(f):
    """Add the hubble flow to velocities in snapshot f"""

    f['vel'] += f['pos'] * H(f)


def comoving_to_physical(ar):
    """Given an array, modify it to be in physical units (remove any
    dependencies on a or aform)."""

    a_power = ar.units._power_of("a")
    aform_power = ar.units._power_of("aform")

    if a_power != 0:
        a = ar.sim.properties['a']
        ar *= a ** a_power
        ar /= units.Unit("a") ** a_power
    if aform_power != 0:
        aform = ar.sim['aform']
        ar *= aform ** aform_power
        ar /= units.Unit("aform")



"""
simdict
=======

This submodule defines an augmented dictionary class
(:class:`SimDict`) for :class:`~pynbody.snapshot.SimSnap` properties
where entries need to be managed e.g.  for defining default entries,
or for ensuring consistency between equivalent properties like
redshift and scalefactor.

By default, a :class:`SimDict` automatically converts between
redshift ('z') and scalefactor ('a') and implements default entries
for cosmological values listed in the [default-cosmology] section of
the `pynbody` configuration files.

Adding further properties
-------------------------

To add further properties use the SimDict.getter and SimDict.setter decorators.
For instance, to add a property 'X_copy' which just reflects the value of the
property 'X', you would use the following code:

.. code-block:: python

 @SimDict.getter
 def X_copy(d) :
     return d['X']

 @SimDict.setter
 def X_copy(d, value) :
     d['X'] = value

"""

import warnings

from . import config

__all__ = ['SimDict']


class SimDict(dict):
    _getters = {}
    _setters = {}

    @staticmethod
    def getter(f):
        """Define a getter function for all SimDicts"""
        SimDict._getters[f.__name__] = f

    @staticmethod
    def setter(f):
        """Define a setter function for all SimDicts"""
        SimDict._setters[f.__name__] = f

    def __getitem__(self, k):
        if k in self:
            return dict.__getitem__(self, k)
        elif k in SimDict._getters:
            return SimDict._getters[k](self)
        else:
            raise KeyError(k)

    def __setitem__(self, k, v):
        if k in SimDict._setters:
            SimDict._setters[k](self, v)
        else:
            dict.__setitem__(self, k, v)


@SimDict.getter
def z(d):
    if d["a"] is None:
        return None
    try:
        return 1.0 / d["a"] - 1.0
    except ZeroDivisionError:
        return None


@SimDict.setter
def z(d, z):
    if z is None:
        d["a"] = None
    else:
        d["a"] = 1.0 / (1.0 + z)


def default_fn(name, value):
    """Return a getter function for the default name/value pair"""
    def f(d):
        warnings.warn("Assuming default value for property '{}'={:.2e}".format(
            name, value), RuntimeWarning)
        d[name] = value
        return value
    f.__name__ = name

    return f

for k in config['default-cosmology']:
    SimDict.getter(default_fn(k, config['default-cosmology'][k]))



"""Interface to OpenMP functions"""

cdef extern from "omp.h" :
     void omp_set_num_threads(int)

cdef extern from "omp.h" :
     int omp_get_max_threads()

cdef extern from "omp.h" :
     int omp_get_num_procs()

def get_threads() :
    """Maps to the OpenMP function omp_get_max_threads()"""
    return omp_get_max_threads()

def set_threads(num) :
    """Maps to the OpenMP function omp_set_num_threads()"""
    omp_set_num_threads(num)

def get_cpus() :
    """Maps to the OpenMP function omp_get_num_procs()"""
    return omp_get_num_procs()



"""

luminosity
==========

Calculates luminosities -- NEEDS DOCUMENTATION

"""

import os

import numpy as np

from .. import filt
from .interpolate import interpolate2d

_cmd_lum_file = os.path.join(os.path.dirname(__file__), "cmdlum.npz")

def use_custom_cmd(path):
    """Use a custom set of stellar populations to calculate magnitudes.

    The path is to a numpy archive with a suitable grid of ages/metallicities and corresponding magnitudes.

    The following script from Stephanie De Beer should help you make a suitable file starting
    from ugriz SSPs downloaded from http://stev.oapd.inaf.it/cgi-bin/cmd.

    import numpy as np

    metals = np.linspace(0.002, 0.05, 25)
    ages = np.logspace(5.67, 10.13, 25)

    mags_bol = np.zeros((len(metals),len(ages)))
    mags_u = np.zeros((len(metals),len(ages)))
    mags_g = np.zeros((len(metals),len(ages)))
    mags_r = np.zeros((len(metals),len(ages)))
    mags_i = np.zeros((len(metals),len(ages)))
    mags_z = np.zeros((len(metals),len(ages)))

    bands = ['bol', 'u', 'g', 'r', 'i', 'z']
    k=2
    for b in bands:
        for x in range(1,26):
            with open('/users/sdebeer/render_stuff/PGSP_files/'+str(x)+'_output.txt', 'r') as f:
                output = f.readlines()
            i = 0
            for line in output:
                magnitudes = line.split()
                if magnitudes[0]=='#':
                    continue
                vars()['mags_'+b][i,x-1]=magnitudes[k]
                i +=1
        k+=1

    np.savez('my_cmd.npz', ages=ages, metals=metals, bol=mags_bol, u=mags_u, g=mags_g, r=mags_r, i=mags_i, z=mags_z)
    """
    global _cmd_lum_file
    _cmd_lum_file = path

def calc_mags(simstars, band='v', cmd_path=None):
    """Calculating visible magnitudes

    Using Padova Simple stellar populations (SSPs) from Girardi
    http://stev.oapd.inaf.it/cgi-bin/cmd
    Marigo+ (2008), Girardi+ (2010)

    pynbody includes a grid of SSP luminosities for many bandpasses for
    various stellar ages and metallicities.  This function linearly
    interpolates to the desired value and returns the value as a magnitude.

    **Usage:**

    >>> import pynbody
    >>> pynbody.analysis.luminosity.calc_mags(h[1].s)

    **Optional keyword arguments:**

       *band* (default='v'): Which observed bandpass magnitude in which
            magnitude should be calculated

       *path* (default=None): Path to the CMD grid. If None, use the
            default or a path specified by use_custom_cmd. For more information
            about generating a custom CMD grid, see use_custom_cmd.

    """

    # find data file in PYTHONPATH
    # data is from http://stev.oapd.inaf.it/cgi-bin/cmd
    # Padova group stellar populations Marigo et al (2008), Girardi et al
    # (2010)
    if cmd_path is not None:
        lums = np.load(cmd_path)
    else:
        lums = np.load(_cmd_lum_file)


    age_star = simstars['age'].in_units('yr')
    # allocate temporary metals that we can play with
    metals = simstars['metals']
    # get values off grid to minmax
    age_star[np.where(age_star < np.min(lums['ages']))] = np.min(lums['ages'])
    age_star[np.where(age_star > np.max(lums['ages']))] = np.max(lums['ages'])
    metals[np.where(metals < np.min(lums['mets']))] = np.min(lums['mets'])
    metals[np.where(metals > np.max(lums['mets']))] = np.max(lums['mets'])

    age_grid = np.log10(lums['ages'])
    met_grid = lums['mets']
    mag_grid = lums[band]

    output_mags = interpolate2d(
        metals, np.log10(age_star), met_grid, age_grid, mag_grid)

    try:
        vals = output_mags - 2.5 * \
            np.log10(simstars['massform'].in_units('Msol'))
    except KeyError:
        vals = output_mags - 2.5 * np.log10(simstars['mass'].in_units('Msol'))

    vals.units = None
    return vals


def halo_mag(sim, band='v'):
    """Calculating halo magnitude

    Calls pynbody.analysis.luminosity.calc_mags for ever star in passed
    in simulation, converts those magnitudes back to luminosities, adds
    those luminosities, then converts that luminosity back to magnitudes,
    which are returned.

    **Usage:**

    >>> import pynbody
    >>> pynbody.analysis.luminosity.halo_mag(h[1].s)

    **Optional keyword arguments:**

       *band* (default='v'): Which observed bandpass magnitude in which
            magnitude should be calculated
    """
    if (len(sim.star) > 0):
        return -2.5 * np.log10(np.sum(10.0 ** (-0.4 * sim.star[band + '_mag'])))
    else:
        return np.nan


def halo_lum(sim, band='v'):
    """Calculating halo luminosiy

    Calls pynbody.analysis.luminosity.calc_mags for every star in passed
    in simulation, converts those magnitudes back to luminosities, adds
    those luminosities, which are returned.  Uses solar magnitudes from
    http://www.ucolick.org/~cnaw/sun.html.

    **Usage:**

    >>> import pynbody
    >>> pynbody.analysis.luminosity.halo_mag(h[1].s)

    **Optional keyword arguments:**

       *band* (default='v'): Which observed bandpass magnitude in which
            magnitude should be calculated
    """
    sun_abs_mag = {'u':5.56,'b':5.45,'v':4.8,'r':4.46,'i':4.1,'j':3.66,
                   'h':3.32,'k':3.28}[band]
    return np.sum(10.0 ** ((sun_abs_mag - sim.star[band + '_mag']) / 2.5))


def half_light_r(sim, band='v', cylindrical=False):
    '''Calculate half light radius

    Calculates entire luminosity of simulation, finds half that, sorts
    stars by distance from halo center, and finds out inside which radius
    the half luminosity is reached.

    If cylindrical is True compute the half light radius as seen from the z-axis.
    '''
    half_l = halo_lum(sim, band=band) * 0.5

    if cylindrical:
        coord = 'rxy'
    else:
        coord = 'r'
    max_high_r = np.max(sim.star[coord])
    test_r = 0.5 * max_high_r
    testrf = filt.LowPass(coord, test_r)
    min_low_r = 0.0
    test_l = halo_lum(sim[testrf], band=band)
    it = 0
    while ((np.abs(test_l - half_l) / half_l) > 0.01):
        it = it + 1
        if (it > 20):
            break

        if (test_l > half_l):
            test_r = 0.5 * (min_low_r + test_r)
        else:
            test_r = (test_r + max_high_r) * 0.5
        testrf = filt.LowPass(coord, test_r)
        test_l = halo_lum(sim[testrf], band=band)

        if (test_l > half_l):
            max_high_r = test_r
        else:
            min_low_r = test_r

    return test_r



import numpy as np
import numpy.testing as npt
import pytest

import pynbody
import pynbody.analysis.hmf as hmf


def test_powspec_static(recwarn):
    f = pynbody.new()
    ps = hmf.PowerSpectrum(f)

    assert any(["Assuming default value" in str(w.message) for w in recwarn.list])
    assert any(["assumes Planck 2018" in str(w.message) for w in recwarn.list])

    assert np.allclose(ps.k.min(), 1.0e-4)
    assert np.allclose(ps.k.max(), 1.0e4)
    assert np.argmax(ps(ps.k)) == 369
    assert np.allclose(ps(ps.k[369]), 24405., rtol=1e-4)

@pytest.mark.parametrize("omegaM, peak_bin, peak_power", [[0.3111, 369, 24405.], [0.25, 355, 39609.]])
def test_powspec_live(recwarn, omegaM, peak_power, peak_bin):
    f = pynbody.new()
    f.properties['omegaM0'] = omegaM

    ps = hmf.PowerSpectrumCAMB(f)

    assert any(["Assuming default value" in str(w.message) for w in recwarn.list])

    assert np.allclose(ps.k.min(), 1.0e-4)
    assert np.allclose(ps.k.max(), 1.0e4)
    assert np.argmax(ps(ps.k)) == peak_bin
    assert np.allclose(ps(ps.k[peak_bin]), peak_power, rtol=1e-4)

def test_get_hmf(recwarn):

    f = pynbody.new()
    m, sig, dn_dlogm = pynbody.analysis.hmf.halo_mass_function(f, log_M_min=10, log_M_max=15, delta_log_M=0.1,
                                                               kern="REEDU", pspec=hmf.PowerSpectrum)

    npt.assert_allclose(m[::10], [1.12201845e+10, 1.12201845e+11, 1.12201845e+12,
                                  1.12201845e+13, 1.12201845e+14], rtol=1e-4)

    npt.assert_allclose(sig[::10], [3.70061772, 2.84212658, 2.08112932, 1.43170708, 0.90729266],
                        rtol=1e-4)

    npt.assert_allclose(dn_dlogm[::10], [4.75228357e-01, 6.23530663e-02, 8.17930180e-03, 9.94758260e-04,
                        8.09997113e-05], rtol=1e-4)

def test_get_empirical_hmf(recwarn):

    f = pynbody.load("testdata/gasoline_ahf/g15784.lr.01024")
    bin_center, bin_counts, err = pynbody.analysis.hmf.simulation_halo_mass_function(f,
                                                                                     log_M_min=10, log_M_max=15,
                                                                                     delta_log_M=1.0)
    npt.assert_allclose(bin_center, [5.5e+10, 5.5e+11, 5.5e+12, 5.5e+13, 5.5e+14])
    npt.assert_allclose(bin_counts, [2.74270242e-03, 9.11568735e-04, 3.83818415e-04, 2.47882726e-04,
          3.99810849e-05])
    npt.assert_allclose(err, [1.48092011e-04, 8.53762344e-05, 5.53994163e-05, 4.45210519e-05,
          1.78800847e-05])
